from datetime import date
from DateTime import DateTime
from pkg_resources import get_distribution
from pkg_resources import parse_version
from plone import api
from plone.app.discussion.interfaces import IDiscussionSettings
from plone.app.layout.navigation.interfaces import INavigationRoot
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.textfield.value import RichTextValue
from plone.dexterity.utils import createContentInContainer
from plone.registry.interfaces import IRegistry
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from plone.restapi.tests.helpers import result_paths
from plone.uuid.interfaces import IMutableUUID
from Products.CMFCore.utils import getToolByName
from zope.component import getUtility
from zope.interface import alsoProvides
from zope.interface import noLongerProvides

import transaction
import unittest


HAS_PLONE_6 = parse_version(
    get_distribution("Products.CMFPlone").version
) >= parse_version("6.0.0a1")


class TestSearchFunctional(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        self.request = self.portal.REQUEST
        self.catalog = getToolByName(self.portal, "portal_catalog")

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        api.user.create(
            email="editor@example.com",
            username="editoruser",
            password="secret",
        )
        api.user.create(
            email="editor@example.com",
            username="localeditor",
            password="secret",
        )

        # /plone/folder
        self.folder = createContentInContainer(
            self.portal, "Folder", id="folder", title="Some Folder"
        )
        api.user.grant_roles(username="editoruser", roles=["Editor"])
        api.user.grant_roles(
            username="localeditor", obj=self.folder, roles=["Editor", "Reader"]
        )

        # /plone/folder/doc
        self.doc = createContentInContainer(
            self.folder,
            "DXTestDocument",
            id="doc",
            title="Lorem Ipsum",
            start=DateTime(1950, 1, 1, 0, 0),
            effective=DateTime(1995, 1, 1, 0, 0),
            expires=DateTime(1999, 1, 1, 0, 0),
            test_int_field=42,
            test_list_field=["Keyword1", "Keyword2", "Keyword3"],
            test_bool_field=True,
            test_richtext_field=RichTextValue(
                raw="<p>Some Text</p>",
                mimeType="text/html",
                outputMimeType="text/html",
            ),
        )
        IMutableUUID(self.doc).set("77779ffa110e45afb1ba502f75f77777")
        self.doc.reindexObject()

        # /plone/folder/other-document
        self.doc2 = createContentInContainer(
            self.folder,
            "DXTestDocument",
            id="other-document",
            title="Other Document",
            description="\xdcbersicht",
            start=DateTime(1975, 1, 1, 0, 0),
            effective=DateTime(2015, 1, 1, 0, 0),
            expires=DateTime(2020, 1, 1, 0, 0),
            test_list_field=["Keyword2", "Keyword3"],
            test_bool_field=False,
        )

        # /plone/folder2
        self.folder2 = createContentInContainer(
            self.portal, "Folder", id="folder2", title="Another Folder"
        )

        # /plone/folder2/doc
        createContentInContainer(
            self.folder2,
            "DXTestDocument",
            id="doc",
            title="Document in second folder",
            start=DateTime(1975, 1, 1, 0, 0),
            effective=DateTime(2015, 1, 1, 0, 0),
            expires=DateTime(2020, 1, 1, 0, 0),
            test_bool_field=False,
        )

        # /plone/doc-outside-folder
        createContentInContainer(
            self.portal,
            "DXTestDocument",
            id="doc-outside-folder",
            title="Doc outside folder",
        )

        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_overall_response_format(self):
        response = self.api_session.get("/@search")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get("Content-Type"), "application/json")

        results = response.json()
        self.assertEqual(
            results["items_total"],
            len(results["items"]),
            "items_total property should match actual item count.",
        )

    def test_search_on_context_constrains_query_by_path(self):
        response = self.api_session.get("/folder/@search")
        self.assertSetEqual(
            {"/plone/folder", "/plone/folder/doc", "/plone/folder/other-document"},
            set(result_paths(response.json())),
        )

    def test_search_in_vhm(self):
        # Install a Virtual Host Monster
        if "virtual_hosting" not in self.app.objectIds():
            # If ZopeLite was imported, we have no default virtual
            # host monster
            from Products.SiteAccess.VirtualHostMonster import (
                manage_addVirtualHostMonster,
            )

            manage_addVirtualHostMonster(self.app, "virtual_hosting")
        transaction.commit()

        # we don't get a result if we do not provide the full physical path
        response = self.api_session.get("/@search?path=/folder")
        self.assertSetEqual(set(), set(result_paths(response.json())))

        # If we go through the VHM will will get results if we only use
        # the part of the path inside the VHM
        vhm_url = "{}/VirtualHostBase/http/plone.org/plone/VirtualHostRoot/{}".format(
            self.app.absolute_url(),
            "@search?path=/folder",
        )
        response = self.api_session.get(vhm_url)
        self.assertSetEqual(
            {"/folder", "/folder/doc", "/folder/other-document"},
            set(result_paths(response.json())),
        )

    def test_search_in_vhm_multiple_paths(self):
        # Install a Virtual Host Monster
        if "virtual_hosting" not in self.app.objectIds():
            # If ZopeLite was imported, we have no default virtual
            # host monster
            from Products.SiteAccess.VirtualHostMonster import (
                manage_addVirtualHostMonster,
            )

            manage_addVirtualHostMonster(self.app, "virtual_hosting")
        transaction.commit()

        # path as a list
        query = {"path": ["/folder", "/folder2"]}

        # If we go through the VHM we will get results for multiple paths
        # if we only use the part of the path inside the VHM
        vhm_url = "{}/VirtualHostBase/http/plone.org/plone/VirtualHostRoot/{}".format(
            self.app.absolute_url(),
            "@search",
        )
        response = self.api_session.get(vhm_url, params=query)
        self.assertSetEqual(
            {
                "/folder",
                "/folder/doc",
                "/folder/other-document",
                "/folder2",
                "/folder2/doc",
            },
            set(result_paths(response.json())),
        )

        # path as a dict with a query list
        query = {"path.query": ["/folder", "/folder2"]}

        # If we go through the VHM we will get results for multiple paths
        # if we only use the part of the path inside the VHM
        vhm_url = "{}/VirtualHostBase/http/plone.org/plone/VirtualHostRoot/{}".format(
            self.app.absolute_url(),
            "@search",
        )
        response = self.api_session.get(vhm_url, params=query)
        self.assertSetEqual(
            {
                "/folder",
                "/folder/doc",
                "/folder/other-document",
                "/folder2",
                "/folder2/doc",
            },
            set(result_paths(response.json())),
        )

    def test_path_gets_prefilled_if_missing_from_path_query_dict(self):
        response = self.api_session.get("/@search?path.depth=1")
        self.assertSetEqual(
            {"/plone/folder", "/plone/folder2", "/plone/doc-outside-folder"},
            set(result_paths(response.json())),
        )

    def test_partial_metadata_retrieval(self):
        query = {
            "SearchableText": "lorem",
            "metadata_fields": ["portal_type", "review_state"],
        }
        response = self.api_session.get("/@search", params=query)

        self.assertDictContainsSubset(
            {
                "@id": self.portal_url + "/folder/doc",
                "title": "Lorem Ipsum",
                "portal_type": "DXTestDocument",
                "review_state": "private",
            },
            response.json()["items"][0],
        )

    def test_full_metadata_retrieval(self):
        query = {"SearchableText": "lorem", "metadata_fields": "_all"}
        response = self.api_session.get("/@search", params=query)

        first_item = response.json()["items"][0]
        self.assertLessEqual(
            {
                "@id": self.portal_url + "/folder/doc",
                "Creator": "test_user_1_",
                "Description": "",
                "EffectiveDate": "None",
                "ExpirationDate": "None",
                "Subject": [],
                "Title": "Lorem Ipsum",
                "Type": "DX Test Document",
                "UID": "77779ffa110e45afb1ba502f75f77777",
                "author_name": None,
                "cmf_uid": None,
                "commentators": [],
                "description": "",
                "effective": "1995-01-01T00:00:00+00:00",
                "end": None,
                "exclude_from_nav": False,
                "expires": "1999-01-01T00:00:00+00:00",
                "getId": "doc",
                "getPath": "/plone/folder/doc",
                "getRemoteUrl": None,
                "getURL": self.portal_url + "/folder/doc",
                "id": "doc",
                "in_response_to": None,
                "is_folderish": False,
                "last_comment_date": None,
                "listCreators": ["test_user_1_"],
                "location": None,
                "portal_type": "DXTestDocument",
                "review_state": "private",
                "start": "1950-01-01T00:00:00+00:00",
                "sync_uid": None,
                "title": "Lorem Ipsum",
                "total_comments": 0,
            }.items(),
            first_item.items(),
        )
        # This value changed in Plone 5.2
        # (Dexterity gained support for getObjSize)
        self.assertIn(first_item["getObjSize"], ("0 KB", "1 KB"))

    def test_full_objects_retrieval(self):
        query = {
            "SearchableText": "lorem",
            "metadata_fields": ["portal_type", "review_state"],
            "fullobjects": True,
        }
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(
            "<p>Some Text</p>",
            response.json()["items"][0]["test_richtext_field"]["data"],
        )
        self.assertEqual(
            self.portal_url + "/folder/doc", response.json()["items"][0]["@id"]
        )

    def test_full_objects_retrieval_discussion(self):
        # Allow discussion
        registry = getUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True
        self.doc.allow_discussion = True

        transaction.commit()

        url = f"{self.doc.absolute_url()}/@comments"
        self.api_session.post(url, json={"text": "comment 1"})
        transaction.commit()

        query = {"portal_type": "Discussion Item", "fullobjects": True}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["items"]), 1)

    def test_full_objects_retrieval_collections(self):
        self.collection = createContentInContainer(
            self.folder, "Collection", id="collection"
        )
        transaction.commit()

        query = {"portal_type": "Collection", "fullobjects": True}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["items"]), 1)

    def test_search_orphan_brain(self):

        # prevent unindex when deleting self.doc
        old__unindexObject = self.doc.__class__.unindexObject
        self.doc.__class__.unindexObject = lambda *args: None
        self.doc.aq_parent.manage_delObjects([self.doc.getId()])
        self.doc.__class__.unindexObject = old__unindexObject
        # doc deleted but still in portal_catalog
        doc_uid = self.doc.UID()
        self.assertFalse(self.doc in self.doc.aq_parent)
        self.assertTrue(self.portal.portal_catalog(UID=doc_uid))
        transaction.commit()

        # query with fullobjects
        query = {"portal_type": "DXTestDocument", "fullobjects": True, "UID": doc_uid}
        response = self.api_session.get("/@search", params=query)
        self.assertEqual(response.status_code, 200, response.content)
        results = response.json()
        self.assertEqual(len(results["items"]), 0)

        # query without fullobjects
        query = {"portal_type": "DXTestDocument", "UID": doc_uid}
        response = self.api_session.get("/@search", params=query)
        self.assertEqual(response.status_code, 200, response.content)
        results = response.json()
        self.assertEqual(len(results["items"]), 1)

    # ZCTextIndex

    def test_fulltext_search(self):
        query = {"SearchableText": "lorem"}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(["/plone/folder/doc"], result_paths(response.json()))

    def test_fulltext_search_with_non_ascii_characters(self):
        query = {"SearchableText": "\xfcbersicht"}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(
            ["/plone/folder/other-document"], result_paths(response.json())
        )

    # KeywordIndex

    def test_keyword_index_str_query(self):
        query = {"test_list_field": "Keyword1"}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(["/plone/folder/doc"], result_paths(response.json()))

    def test_keyword_index_str_query_or(self):
        query = {"test_list_field": ["Keyword2", "Keyword3"]}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(
            ["/plone/folder/doc", "/plone/folder/other-document"],
            result_paths(response.json()),
        )

    def test_keyword_index_str_query_and(self):
        query = {
            "test_list_field.query": ["Keyword1", "Keyword2"],
            "test_list_field.operator": "and",
        }
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(["/plone/folder/doc"], result_paths(response.json()))

    # BooleanIndex

    def test_boolean_index_query(self):
        query = {"test_bool_field": True, "portal_type": "DXTestDocument"}
        response = self.api_session.get("/folder/@search", params=query)
        self.assertEqual(["/plone/folder/doc"], result_paths(response.json()))

        query = {"test_bool_field": False, "portal_type": "DXTestDocument"}
        response = self.api_session.get("/folder/@search", params=query)
        self.assertEqual(
            ["/plone/folder/other-document"], result_paths(response.json())
        )

    # FieldIndex

    def test_field_index_int_query(self):
        query = {"test_int_field:int": 42}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(["/plone/folder/doc"], result_paths(response.json()))

    def test_field_index_int_range_query(self):
        query = {
            "test_int_field.query:int": [41, 43],
            "test_int_field.range": "min:max",
        }
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(["/plone/folder/doc"], result_paths(response.json()))

    # ExtendedPathIndex

    def test_extended_path_index_query(self):
        query = {"path": "/".join(self.folder.getPhysicalPath())}

        response = self.api_session.get("/@search", params=query)

        self.assertEqual(
            sorted(
                [
                    "/plone/folder",
                    "/plone/folder/doc",
                    "/plone/folder/other-document",
                ]
            ),
            sorted(result_paths(response.json())),
        )

    def test_extended_path_index_query_multiple(self):
        # path as a list
        query = {
            "path": [
                "/".join(self.folder.getPhysicalPath()),
                "/".join(self.folder2.getPhysicalPath()),
            ]
        }
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(
            sorted(
                [
                    "/plone/folder",
                    "/plone/folder/doc",
                    "/plone/folder/other-document",
                    "/plone/folder2",
                    "/plone/folder2/doc",
                ]
            ),
            sorted(result_paths(response.json())),
        )

        # path as a dict with a query list
        query = {
            "path.query": [
                "/".join(self.folder.getPhysicalPath()),
                "/".join(self.folder2.getPhysicalPath()),
            ]
        }
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(
            sorted(
                [
                    "/plone/folder",
                    "/plone/folder/doc",
                    "/plone/folder/other-document",
                    "/plone/folder2",
                    "/plone/folder2/doc",
                ]
            ),
            sorted(result_paths(response.json())),
        )

    def test_extended_path_index_depth_limiting(self):
        lvl1 = createContentInContainer(self.portal, "Folder", id="lvl1")
        lvl2 = createContentInContainer(lvl1, "Folder", id="lvl2")
        createContentInContainer(lvl2, "Folder", id="lvl3")
        transaction.commit()

        path = "/plone/lvl1"

        # Depth 0 - only object identified by path
        query = {"path.query": path, "path.depth": 0}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(["/plone/lvl1"], result_paths(response.json()))

        # Depth 1 - immediate children
        query = {"path.query": path, "path.depth": 1}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(["/plone/lvl1/lvl2"], result_paths(response.json()))

        # No depth - object itself and all children
        query = {"path": path}
        response = self.api_session.get("/@search", params=query)

        self.assertSetEqual(
            {"/plone/lvl1", "/plone/lvl1/lvl2", "/plone/lvl1/lvl2/lvl3"},
            set(result_paths(response.json())),
        )

    # DateIndex

    def test_date_index_query(self):
        query = {"start": date(1950, 1, 1).isoformat()}
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(["/plone/folder/doc"], result_paths(response.json()))

    def test_date_index_ranged_query(self):
        query = {
            "start.query": [date(1949, 1, 1).isoformat(), date(1951, 1, 1).isoformat()],
            "start.range": "min:max",
        }
        response = self.api_session.get("/@search", params=query)

        self.assertEqual(["/plone/folder/doc"], result_paths(response.json()))

    # DateRangeIndex

    def test_date_range_index_query(self):
        query = {"effectiveRange": date(1997, 1, 1).isoformat()}
        response = self.api_session.get("/folder/@search", params=query)

        self.assertEqual(2, len(result_paths(response.json())))
        self.assertTrue("/plone/folder" in result_paths(response.json()))
        self.assertTrue("/plone/folder/doc" in result_paths(response.json()))

    # DateRecurringIndex

    def test_date_recurring_index_query(self):
        from datetime import datetime

        createContentInContainer(
            self.folder,
            "Event",
            id="event",
            title="Event",
            start=datetime(2013, 1, 1, 0, 0),
            end=datetime(2013, 1, 1, 23, 59),
            whole_day=True,
            recurrence="FREQ=DAILY;COUNT=10;INTERVAL=2",
            timezone="UTC",
        )
        import transaction

        transaction.commit()

        # First occurrence
        query = {"start": date(2013, 1, 1).isoformat()}
        response = self.api_session.get("/folder/@search", params=query)

        self.assertEqual(["/plone/folder/event"], result_paths(response.json()))

        # No event that day
        query = {"start": date(2013, 1, 2).isoformat()}
        response = self.api_session.get("/folder/@search", params=query)

        self.assertEqual([], result_paths(response.json()))

        # Second occurrence
        query = {"start": date(2013, 1, 3).isoformat()}
        response = self.api_session.get("/folder/@search", params=query)

        self.assertEqual(["/plone/folder/event"], result_paths(response.json()))

        # Ranged query
        query = {
            "start.query": [date(2013, 1, 1).isoformat(), date(2013, 1, 5).isoformat()],
            "start.range": "min:max",
        }
        response = self.api_session.get("/folder/@search", params=query)

        self.assertEqual(["/plone/folder/event"], result_paths(response.json()))

    # UUIDIndex

    def test_uuid_index_query(self):
        IMutableUUID(self.doc).set("7777a074cb4240d08c9a129e3a837777")
        self.doc.reindexObject()
        transaction.commit()

        query = {"UID": "7777a074cb4240d08c9a129e3a837777"}
        response = self.api_session.get("/@search", params=query)
        self.assertEqual(["/plone/folder/doc"], result_paths(response.json()))

    def test_respect_access_inactive_permission(self):
        # admin can see everything
        response = self.api_session.get("/@search", params={}).json()
        if HAS_PLONE_6:
            # Since Plone 6 the Plone site is indexed ...
            self.assertEqual(response["items_total"], 7)
        else:
            # ... before it was not
            self.assertEqual(response["items_total"], 6)
        response = self.api_session.get(
            "/@search", params={"Title": "Lorem Ipsum"}
        ).json()
        self.assertEqual(response["items_total"], 1)

        # not admin users can't see expired items
        self.api_session.auth = ("editoruser", "secret")

        response = self.api_session.get("/@search", params={}).json()
        if HAS_PLONE_6:
            # Since Plone 6 the Plone site is indexed ...
            self.assertEqual(response["items_total"], 4)
        else:
            # ... before it was not
            self.assertEqual(response["items_total"], 3)
        response = self.api_session.get(
            "/@search", params={"Title": "Lorem Ipsum"}
        ).json()
        self.assertEqual(response["items_total"], 0)

        # now grant permission to Editor to access inactive content
        self.portal.manage_permission(
            "Access inactive portal content", roles=["Manager", "Editor"]
        )
        transaction.commit()

        #  portal-enabled Editor can see expired contents
        response = self.api_session.get("/@search", params={}).json()
        if HAS_PLONE_6:
            # Since Plone 6 the Plone site is indexed ...
            self.assertEqual(response["items_total"], 7)
        else:
            # ... before it was not
            self.assertEqual(response["items_total"], 6)
        response = self.api_session.get(
            "/@search", params={"Title": "Lorem Ipsum"}
        ).json()
        self.assertEqual(response["items_total"], 1)

        # local-enabled Editor can only access expired contents inside folder
        self.api_session.auth = ("localeditor", "secret")
        response = self.api_session.get("/@search", params={}).json()
        if HAS_PLONE_6:
            # Since Plone 6 the Plone site is indexed ...
            self.assertEqual(response["items_total"], 2)
        else:
            # ... before it was not
            self.assertEqual(response["items_total"], 1)
        response = self.api_session.get(
            "/@search", params={"path": "/plone/folder"}
        ).json()

        self.assertEqual(response["items_total"], 3)
        response = self.api_session.get(
            "/@search", params={"Title": "Lorem Ipsum"}
        ).json()
        self.assertEqual(response["items_total"], 0)
        response = self.api_session.get(
            "/@search",
            params={"Title": "Lorem Ipsum", "path": "/plone/folder"},
        ).json()
        self.assertEqual(response["items_total"], 1)

    def test_search_use_site_search_settings_for_types(self):
        response = self.api_session.get(
            "/@search", params={"use_site_search_settings": 1}
        ).json()
        types = {item["@type"] for item in response["items"]}

        self.assertEqual(set(types), {"Folder", "DXTestDocument"})

        registry = getUtility(IRegistry)
        from Products.CMFPlone.interfaces import ISearchSchema

        search_settings = registry.forInterface(ISearchSchema, prefix="plone")
        old = search_settings.types_not_searched
        search_settings.types_not_searched += ("DXTestDocument",)
        transaction.commit()

        response = self.api_session.get(
            "/@search", params={"use_site_search_settings": 1}
        ).json()
        types = {item["@type"] for item in response["items"]}

        self.assertEqual(set(types), {"Folder"})
        search_settings.types_not_searched = old
        transaction.commit()

    def test_search_use_site_search_settings_for_default_sort_order(self):
        response = self.api_session.get(
            "/@search", params={"use_site_search_settings": 1}
        ).json()
        titles = [
            "Some Folder",
            "Lorem Ipsum",
            "Other Document",
            "Another Folder",
            "Document in second folder",
            "Doc outside folder",
        ]
        self.assertEqual([item["title"] for item in response["items"]], titles)

        response = self.api_session.get(
            "/@search", params={"use_site_search_settings": 1, "sort_on": "effective"}
        ).json()
        self.assertEqual(
            [item["title"] for item in response["items"]][0],
            "Other Document",
        )

    def test_search_use_site_search_settings_with_navigation_root(self):

        alsoProvides(self.folder, INavigationRoot)
        transaction.commit()

        response = self.api_session.get(
            "/folder/@search", params={"use_site_search_settings": 1}
        ).json()
        titles = ["Some Folder", "Lorem Ipsum", "Other Document"]
        self.assertEqual([item["title"] for item in response["items"]], titles)

        noLongerProvides(self.folder, INavigationRoot)
        transaction.commit()

    def test_search_use_site_search_settings_with_navigation_root_and_vhm(self):

        if "virtual_hosting" not in self.app.objectIds():
            # If ZopeLite was imported, we have no default virtual
            # host monster
            from Products.SiteAccess.VirtualHostMonster import (
                manage_addVirtualHostMonster,
            )

            manage_addVirtualHostMonster(self.app, "virtual_hosting")
        alsoProvides(self.folder, INavigationRoot)
        transaction.commit()

        vhm_url = "{}/VirtualHostBase/http/plone.org/plone/VirtualHostRoot/{}".format(
            self.app.absolute_url(),
            "/folder/@search",
        )
        response = self.api_session.get(
            vhm_url, params={"use_site_search_settings": 1, "path": "/folder"}
        ).json()
        titles = ["Some Folder", "Lorem Ipsum", "Other Document"]
        self.assertEqual([item["title"] for item in response["items"]], titles)

        noLongerProvides(self.folder, INavigationRoot)
        transaction.commit()

    def test_search_use_site_search_settings_with_vhm(self):

        if "virtual_hosting" not in self.app.objectIds():
            # If ZopeLite was imported, we have no default virtual
            # host monster
            from Products.SiteAccess.VirtualHostMonster import (
                manage_addVirtualHostMonster,
            )

            manage_addVirtualHostMonster(self.app, "virtual_hosting")
        transaction.commit()

        vhm_url = "{}/VirtualHostBase/http/plone.org/plone/VirtualHostRoot/{}".format(
            self.app.absolute_url(),
            "/@search",
        )
        response = self.api_session.get(
            vhm_url, params={"use_site_search_settings": 1, "path": "/"}
        ).json()
        titles = sorted(
            [
                "Another Folder",
                "Doc outside folder",
                "Document in second folder",
                "Lorem Ipsum",
                "Other Document",
                "Some Folder",
            ]
        )
        self.assertEqual(sorted([item["title"] for item in response["items"]]), titles)

        noLongerProvides(self.folder, INavigationRoot)
        transaction.commit()



# pylint: disable=E0211, W0221
# E0211: Method has no argument
# W0221: Arguments number differs from overridden '__call__' method

from zope.interface import Attribute
from zope.interface import Interface
from zope.publisher.interfaces.browser import IDefaultBrowserLayer


class IPloneRestapiLayer(IDefaultBrowserLayer):
    """Marker interface that defines a browser layer."""


class ISerializeToJson(Interface):
    """Adapter to serialize a Dexterity object into a JSON object."""


class ISerializeToJsonSummary(Interface):
    """Adapter to serialize an object into a JSON compatible summary that
    contains only the most basic information.
    """


class IJsonCompatible(Interface):
    """Convert a value to a JSON compatible data structure."""


class IContextawareJsonCompatible(IJsonCompatible):
    """Convert a value to a JSON compatible data structure, using a context."""

    def __init__(value, context):
        """Adapts value and a context"""


class IFieldSerializer(Interface):
    """The field serializer multi adapter serializes the field value into
    JSON compatible python data.
    """

    def __init__(field, context, request):
        """Adapts field, context and request."""

    def __call__():
        """Returns JSON compatible python data."""


class IPrimaryFieldTarget(Interface):
    """Return a URL to direct the user to if this is the primary field.
    Useful e.g. if you want to redirect certain users to a download url
    instead of the item's view.
    """

    def __init__(field, context, request):
        """Adapts field, context and request."""

    def __call__():
        """Returns a URL."""


class IObjectPrimaryFieldTarget(Interface):
    """Return a URL to direct the user to if the object has a primary field
    that provides an IPrimaryFieldTarget.
    """

    def __init__(field, context, request):
        """Adapts field, context and request."""

    def __call__():
        """Returns a URL."""


class IDeserializeFromJson(Interface):
    """An adapter to deserialize a JSON object into an object in Plone."""


class IFieldDeserializer(Interface):
    """An adapter to deserialize a JSON value into a field value."""

    def __init__(field, context, request):
        """Adapts a field, it's context and the request."""

    def __call__(value):
        """Convert the provided JSON value to a field value."""


class IBlockFieldDeserializationTransformer(Interface):
    """Convert/adjust raw block deserialized value into block value."""

    block_type = Attribute(
        "A string with the type of block, the @type from " "the block value"
    )
    order = Attribute(
        "A number used in sorting value transformers. " "Smaller is executed first"
    )
    disabled = Attribute("Boolean that disables the transformer if required")

    def __init__(field, context, request):
        """Adapts context and the request."""

    def __call__(value):
        """Convert the provided raw Python value to a block value."""


class IBlockFieldSerializationTransformer(Interface):
    """Transform block value before final JSON serialization"""

    block_type = Attribute(
        "A string with the type of block, the @type from " "the block value"
    )
    order = Attribute(
        "A number used in sorting value transformers for the "
        "same block. Smaller is executed first"
    )
    disabled = Attribute("Boolean that disables the transformer if required")

    def __init__(field, context, request):
        """Adapts context and the request."""

    def __call__(value):
        """Convert the provided raw Python value to a block value."""


class IExpandableElement(Interface):
    """A named adapter that deserializes an element in expanded or collapsed
    form.
    """

    def __call__(expand=False):
        """ """


class IZCatalogCompatibleQuery(Interface):
    """A multi adapter responsible for converting a catalog query provided as
    a Python dictionary, but with possibly incorrect value types, to a
    Python dictionary that can be passed directly to catalog.searchResults().

    Values (query values or query options) that can't be serialized in JSON
    (like datetimes) or a query string (any type other than string) must be
    converted back by this adapter, by delegating that job to an
    IIndexQueryParser for each of the queried indexes.
    """

    global_query_params = Attribute(
        "A mapping of query-wide parameters (like 'sort_on') to their data "
        "type. These need to be treated separately from indexes."
    )

    def __init__(context, request):
        """Adapts context and request."""

    def __call__(query):
        """Returns a ZCatalog compatible query (Python dictionary)."""


class IIndexQueryParser(Interface):
    """A multi adapter responsible for deserializing values in catalog query
    options for a particular index type.

    The main issue here is typing of query values and query options: ZCatalog
    expects any values in queries to have the proper type, and fails
    unpredictably otherwise.

    Because we can't avoid losing (some of) this typing information when
    submitting queries to the API via a HTTP query string, we need to
    reconstruct it based on what the index that is queried expects.

    This adapter therefore needs to know what data types the adapted index
    expects, and turn any values (query values or query options) back into the
    proper data type.
    """

    query_value_type = Attribute(
        "The data type of the query value for queries against this index. "
        "The query value may also be a sequence of values of that type."
    )

    query_options = Attribute(
        "A mapping of query options this index type supports to their type."
    )

    def __init__(index, context, request):
        """Adapts a ZCatalog index, context and request."""

    def parse(idx_query):
        """Takes a query against a single index (the value part of a
        {'index_name': idx_query} pair).

        `idx_query` can therefore be
          - a single string value
          - a list of string values
          - a dictionary with one or more query options, among them the actual
            query value identified by the 'query' key

        Returns a transformed `idx_query` whose query options and query values
        have been reconstructed to the proper data types that the adapted
        index expects.
        """


class IBlockSearchableText(Interface):
    """Allow blocks to provide text for the SearchableText index

    Register as a named adapter, where the name is the block @type
    """

    def __init__(field, context, request):
        """Adapts a context and the request."""

    def __call__(value):
        """Extract text from the block value. Returns text"""


class IJSONSummarySerializerMetadata(Interface):
    """Configure JSONSummary serializer."""

    def default_metadata():
        """Returns a set with default metadata to be serialized."""

    def field_accessors():
        """Returns a dictionary with field accessors to be used during serialization."""

    def non_metadata_attributes():
        """Returns a set with non metadata attributes."""

    def blocklisted_attributes():
        """Returns a set with attributes blocked during serialization."""



from datetime import datetime as dt
from plone.app.layout.viewlets.content import ContentHistoryViewlet
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.serializer.converters import json_compatible
from plone.restapi.services import Service
from Products.CMFPlone.utils import safe_unicode
from zope.component import queryMultiAdapter
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class HistoryGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.version = None

    def publishTraverse(self, request, name):
        self.version = name
        return self

    def reply(self):
        # Traverse to historical version
        if self.version:
            serializer = queryMultiAdapter(
                (self.context, self.request), ISerializeToJson
            )
            data = serializer(version=self.version)
            return data

        # Listing historical data
        content_history_viewlet = ContentHistoryViewlet(
            self.context, self.request, None, None
        )
        site_url = getSite().absolute_url()
        content_history_viewlet.navigation_root_url = site_url
        content_history_viewlet.site_url = site_url
        history = content_history_viewlet.fullHistory()
        if history is None:
            history = []

        unwanted_keys = [
            "diff_current_url",
            "diff_previous_url",
            "preview_url",
            "actor_home",
            "actorid",
            "revert_url",
            "version_id",
        ]

        for item in history:
            item["actor"] = {
                "@id": "{}/@users/{}".format(site_url, item["actorid"]),
                "id": item["actorid"],
                "fullname": item["actor"].get("fullname"),
                "username": item["actor"].get("username"),
            }

            if item["type"] == "versioning":
                item["version"] = item["version_id"]
                item["@id"] = "{}/@history/{}".format(
                    self.context.absolute_url(), item["version"]
                )

                # If a revert_url is present, then CMFEditions has checked our
                # permissions.
                item["may_revert"] = bool(item.get("revert_url"))

            # Versioning entries use a timestamp,
            # workflow ISO formatted string
            if not isinstance(item["time"], str):
                item["time"] = dt.fromtimestamp(item["time"]).isoformat()

            # The create event has an empty 'action', but we like it to say
            # 'Create', alike the transition_title
            if item["action"] is None:
                item["action"] = "Create"

            # We want action, state and transition names translated
            if "state_title" in item:
                item["state_title"] = self.context.translate(
                    safe_unicode(item["state_title"]), context=self.request
                )

            if "transition_title" in item:
                item["transition_title"] = self.context.translate(
                    safe_unicode(item["transition_title"]), context=self.request
                )

            if "action" in item:
                item["action"] = self.context.translate(
                    safe_unicode(item["action"]), context=self.request
                )

            # clean up
            for key in unwanted_keys:
                if key in item:
                    del item[key]

        return json_compatible(history)



from plone.restapi.controlpanels.interfaces import IDexterityTypesControlpanel
from plone.restapi.deserializer import json_body
from plone.restapi.deserializer.controlpanels import ControlpanelDeserializeFromJson
from plone.restapi.interfaces import IDeserializeFromJson
from zope.component import adapter, queryMultiAdapter
from zope.interface import implementer


@implementer(IDeserializeFromJson)
@adapter(IDexterityTypesControlpanel)
class DexterityTypesControlpanelDeserializeFromJson(ControlpanelDeserializeFromJson):
    def deserialize_item(self, proxy):
        data = json_body(self.request)
        overview = queryMultiAdapter((proxy, self.request), name="overview")
        overview.form_instance.applyChanges(data)
        behaviors = queryMultiAdapter((proxy, self.request), name="behaviors")
        behaviors.form_instance.applyChanges(data)

    def __call__(self, item=None):
        if item is not None:
            return self.deserialize_item(item)
        return super().__call__()



from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession

import unittest


class TestRolesGet(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

    def tearDown(self):
        self.api_session.close()

    def test_roles_endpoint_lists_roles(self):
        response = self.api_session.get("/@roles")

        expected = (
            {
                "@id": self.portal_url + "/@roles/Contributor",
                "@type": "role",
                "id": "Contributor",
                "title": "Contributor",
            },
            {
                "@id": self.portal_url + "/@roles/Editor",
                "@type": "role",
                "id": "Editor",
                "title": "Editor",
            },
            {
                "@id": self.portal_url + "/@roles/Member",
                "@type": "role",
                "id": "Member",
                "title": "Member",
            },
            {
                "@id": self.portal_url + "/@roles/Reader",
                "@type": "role",
                "id": "Reader",
                "title": "Reader",
            },
            {
                "@id": self.portal_url + "/@roles/Reviewer",
                "@type": "role",
                "id": "Reviewer",
                "title": "Reviewer",
            },
            {
                "@id": self.portal_url + "/@roles/Site Administrator",
                "@type": "role",
                "id": "Site Administrator",
                "title": "Site Administrator",
            },
            {
                "@id": self.portal_url + "/@roles/Manager",
                "@type": "role",
                "id": "Manager",
                "title": "Manager",
            },
        )
        result = response.json()
        self.assertEqual(len(expected), len(result))
        for item in result:
            self.assertIn(item, expected)

    def test_roles_endpoint_translates_role_titles(self):
        self.api_session.headers.update({"Accept-Language": "de"})
        response = self.api_session.get("/@roles")
        # One of the roles has changed translation in German.
        # Reviewer used to be 'Ver\xf6ffentlichen', but is now simply Reviewer.
        titles = {item["title"] for item in response.json()}
        options = {"Ver\xf6ffentlichen", "Reviewer"}
        # One of the options must match:
        self.assertTrue(titles.intersection(options))
        # Discard them:
        titles = titles.difference(options)
        self.assertEqual(
            {
                "Hinzuf\xfcgen",
                "Bearbeiten",
                "Benutzer",
                "Ansehen",
                "Website-Administrator",
                "Verwalten",
            },
            titles,
        )



from importlib import import_module
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.schema import SCHEMA_CACHE
from plone.restapi.interfaces import IDeserializeFromJson
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.restapi.tests.mixin_ordering import OrderingMixin
from zope.component import getMultiAdapter
from zope.component import queryUtility

import json
import unittest

HAS_PLONE_6 = getattr(
    import_module("Products.CMFPlone.factory"), "PLONE60MARKER", False
)


class TestDXContentDeserializer(unittest.TestCase, OrderingMixin):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        # ordering setup
        self.folder = self.portal

        for x in range(1, 10):
            self.folder.invokeFactory(
                "Document", id="doc" + str(x), title="Test doc " + str(x)
            )

    def deserialize(self, body="{}", validate_all=False, context=None):
        context = context or self.portal
        self.request["BODY"] = body
        deserializer = getMultiAdapter((context, self.request), IDeserializeFromJson)
        return deserializer(validate_all=validate_all)

    def test_set_layout(self):
        current_layout = self.portal.getLayout()
        self.assertNotEqual(current_layout, "my_new_layout")
        self.deserialize(body='{"layout": "my_new_layout"}')
        self.assertEqual("my_new_layout", self.portal.getLayout())


class TestSiteRootDeserializer(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.portal.invokeFactory(
            "Document",
            id="doc1",
        )

        # Enable volto.blocks if Plone Site is DX
        fti = queryUtility(IDexterityFTI, name="Plone Site")
        if fti is not None:
            behavior_list = [a for a in fti.behaviors]
            behavior_list.append("volto.blocks")
            fti.behaviors = tuple(behavior_list)
            # Invalidating the cache is required for the FTI to be applied
            # on the existing object
            SCHEMA_CACHE.invalidate("Plone Site")

    def deserialize(self, body="{}", validate_all=False, context=None):
        context = context or self.portal
        self.request["BODY"] = body
        deserializer = getMultiAdapter((context, self.request), IDeserializeFromJson)
        return deserializer(validate_all=validate_all)

    @unittest.skipIf(
        HAS_PLONE_6,
        "This test is only intended to run for Plone 5 and the blocks behavior site root hack enabled",
    )
    def test_opt_in_blocks_deserializer_plone5(self):
        blocks = {
            "0358abe2-b4f1-463d-a279-a63ea80daf19": {"@type": "description"},
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
        }
        blocks_layout = {
            "items": [
                "07c273fc-8bfc-4e7d-a327-d513e5a945bb",
                "0358abe2-b4f1-463d-a279-a63ea80daf19",
            ]
        }

        self.deserialize(
            body='{{"blocks": {}, "blocks_layout": {}}}'.format(
                json.dumps(blocks), json.dumps(blocks_layout)
            )
        )

        self.assertEqual(blocks, json.loads(self.portal.blocks))
        self.assertEqual(blocks_layout, json.loads(self.portal.blocks_layout))

    @unittest.skipIf(
        HAS_PLONE_6,
        "This test is only intended to run for Plone 5 and the blocks behavior site root hack enabled",
    )
    def test_resolveuids_blocks_deserializer_plone5(self):
        blocks = {
            "0358abe2-b4f1-463d-a279-a63ea80daf19": {
                "@type": "foo",
                "url": self.portal.doc1.absolute_url(),
            },
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
        }
        blocks_layout = {
            "items": [
                "07c273fc-8bfc-4e7d-a327-d513e5a945bb",
                "0358abe2-b4f1-463d-a279-a63ea80daf19",
            ]
        }

        self.deserialize(
            body='{{"blocks": {}, "blocks_layout": {}}}'.format(
                json.dumps(blocks), json.dumps(blocks_layout)
            )
        )

        values = json.loads(self.portal.blocks)
        self.assertEqual(
            values["0358abe2-b4f1-463d-a279-a63ea80daf19"]["url"],
            f"resolveuid/{self.portal.doc1.UID()}",
        )

    @unittest.skipIf(
        not HAS_PLONE_6,
        "This test is only intended to run for Plone 6 and DX site root enabled",
    )
    def test_opt_in_blocks_deserializer(self):
        blocks = {
            "0358abe2-b4f1-463d-a279-a63ea80daf19": {"@type": "description"},
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
        }
        blocks_layout = {
            "items": [
                "07c273fc-8bfc-4e7d-a327-d513e5a945bb",
                "0358abe2-b4f1-463d-a279-a63ea80daf19",
            ]
        }

        self.deserialize(
            body='{{"blocks": {}, "blocks_layout": {}}}'.format(
                json.dumps(blocks), json.dumps(blocks_layout)
            )
        )

        self.assertEqual(blocks, self.portal.blocks)
        self.assertEqual(blocks_layout, self.portal.blocks_layout)

    @unittest.skipIf(
        not HAS_PLONE_6,
        "This test is only intended to run for Plone 6 and DX site root enabled",
    )
    def test_resolveuids_blocks_deserializer(self):
        blocks = {
            "0358abe2-b4f1-463d-a279-a63ea80daf19": {
                "@type": "foo",
                "url": self.portal.doc1.absolute_url(),
            },
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
        }
        blocks_layout = {
            "items": [
                "07c273fc-8bfc-4e7d-a327-d513e5a945bb",
                "0358abe2-b4f1-463d-a279-a63ea80daf19",
            ]
        }

        self.deserialize(
            body='{{"blocks": {}, "blocks_layout": {}}}'.format(
                json.dumps(blocks), json.dumps(blocks_layout)
            )
        )

        values = self.portal.blocks
        self.assertEqual(
            values["0358abe2-b4f1-463d-a279-a63ea80daf19"]["url"],
            f"resolveuid/{self.portal.doc1.UID()}",
        )



from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from zExceptions import BadRequest
from zope.component import queryMultiAdapter
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


DEFAULT_SEARCH_RESULTS_LIMIT = 25


@implementer(IPublishTraverse)
class GroupsGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []
        self.query = self.request.form.copy()

    def publishTraverse(self, request, name):
        # Consume any path segments after /@users as parameters
        self.params.append(name)
        return self

    @property
    def _get_group_id(self):
        if len(self.params) != 1:
            raise Exception("Must supply exactly one parameter (group id)")
        return self.params[0]

    def _get_group(self, group_id):
        portal = getSite()
        portal_groups = getToolByName(portal, "portal_groups")
        return portal_groups.getGroupById(group_id)

    def _get_groups(self):
        portal = getSite()
        portal_groups = getToolByName(portal, "portal_groups")
        return portal_groups.listGroups()

    def _get_filtered_groups(self, query, limit):
        portal = getSite()
        portal_groups = getToolByName(portal, "portal_groups")
        results = portal_groups.searchGroups(id=query, max_results=limit)
        return [portal_groups.getGroupById(group["groupid"]) for group in results]

    def reply(self):
        if len(self.query) > 0 and len(self.params) == 0:
            query = self.query.get("query", "")
            limit = self.query.get("limit", DEFAULT_SEARCH_RESULTS_LIMIT)
            if query:
                groups = self._get_filtered_groups(query, limit)
                result = []
                for group in groups:
                    serializer = queryMultiAdapter(
                        (group, self.request), ISerializeToJsonSummary
                    )
                    result.append(serializer())
                return result
            else:
                raise BadRequest("Query string supplied is not valid")

        if len(self.params) == 0:
            result = []
            for group in self._get_groups():
                serializer = queryMultiAdapter(
                    (group, self.request), ISerializeToJsonSummary
                )
                result.append(serializer())
            return result
        # we retrieve the user on the user id not the username
        group = self._get_group(self._get_group_id)
        if not group:
            self.request.response.setStatus(404)
            return
        serializer = queryMultiAdapter((group, self.request), ISerializeToJson)
        return serializer()



from zope.interface import Attribute
from zope.interface import Interface


class IControlpanel(Interface):
    __name__ = Attribute("Name of the controlpanel in the URL")
    title = Attribute("Title of this controlpanel")
    group = Attribute("Group name of this controlpanel")
    schema = Attribute("Registry schema of this controlpanel")

    configlet_id = Attribute("Id of the configlet, e.g. MailHost")
    configlet_category_id = Attribute(
        "Category of the configlet, e.g. plone-general"
    )  # noqa

    def add(names):
        """Create controlpanel children by names"""

    def get(names):
        """Read controlpanel children by names"""

    def update(names):
        """Update controlpanel children by names"""

    def delete(names):
        """Remove controlpanel children by names"""


class IDexterityTypesControlpanel(IControlpanel):
    """Dexterity Types Control panel"""



from base64 import b64encode
from plone import api
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.restapi.testing import RelativeSession
from zope.event import notify
from ZPublisher.pubevents import PubStart

import transaction
import unittest


class TestCopyMove(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.doc1 = self.portal[
            self.portal.invokeFactory("Document", id="doc1", title="My Document")
        ]
        self.folder1 = self.portal[
            self.portal.invokeFactory("Folder", id="folder1", title="My Folder")
        ]

    def traverse(self, path="/plone", accept="application/json", method="GET"):
        request = self.layer["request"]
        request.environ["PATH_INFO"] = path
        request.environ["PATH_TRANSLATED"] = path
        request.environ["HTTP_ACCEPT"] = accept
        request.environ["REQUEST_METHOD"] = method
        auth = f"{SITE_OWNER_NAME}:{SITE_OWNER_PASSWORD}"
        request._auth = "Basic %s" % b64encode(auth.encode("utf8")).decode("utf8")
        notify(PubStart(request))
        return request.traverse(path)

    def test_get_object_by_url(self):
        service = self.traverse("/plone/@copy", method="POST")
        obj = service.get_object(self.doc1.absolute_url())

        self.assertEqual(self.doc1, obj)

    def test_get_object_by_path(self):
        service = self.traverse("/plone/@copy", method="POST")
        obj = service.get_object("/doc1")

        self.assertEqual(self.doc1, obj)

    def test_get_object_by_uid(self):
        service = self.traverse("/plone/@copy", method="POST")
        obj = service.get_object(self.doc1.UID())

        self.assertEqual(self.doc1, obj)


class TestCopyMoveFunctional(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.doc1 = self.portal[
            self.portal.invokeFactory("Document", id="doc1", title="My Document")
        ]
        self.doc2 = self.portal[
            self.portal.invokeFactory("Document", id="doc2", title="My Document")
        ]
        self.folder1 = self.portal[
            self.portal.invokeFactory("Folder", id="folder1", title="My Folder")
        ]

        api.user.create(
            email="memberuser@example.com", username="memberuser", password="secret"
        )

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_copy_single_object(self):
        response = self.api_session.post(
            "/@copy", json={"source": self.doc1.absolute_url()}
        )
        transaction.commit()

        self.assertEqual(response.status_code, 200)
        self.assertIn("copy_of_doc1", self.portal.objectIds())

    def test_move_single_object(self):
        response = self.api_session.post(
            "/folder1/@move", json={"source": self.doc1.absolute_url()}
        )
        transaction.commit()

        self.assertEqual(response.status_code, 200)
        self.assertIn("doc1", self.folder1.objectIds())
        self.assertNotIn("doc1", self.portal.objectIds())

    def test_move_multiple_objects(self):
        response = self.api_session.post(
            "/folder1/@move",
            json={"source": [self.doc1.absolute_url(), self.doc2.absolute_url()]},
        )
        self.assertEqual(response.status_code, 200)
        transaction.commit()

        self.assertIn("doc1", self.folder1.objectIds())
        self.assertIn("doc2", self.folder1.objectIds())
        self.assertNotIn("doc1", self.portal.objectIds())
        self.assertNotIn("doc2", self.portal.objectIds())

    def test_copy_without_source_raises_400(self):
        response = self.api_session.post("/folder1/@copy")
        self.assertEqual(response.status_code, 400)

    def test_copy_not_existing_object(self):
        response = self.api_session.post("/@copy", json={"source": "does-not-exist"})

        self.assertEqual(response.status_code, 200)
        self.assertEqual([], response.json())

    def test_copy_multiple_objects(self):
        response = self.api_session.post(
            "/@copy",
            json={"source": [self.doc1.absolute_url(), self.doc2.absolute_url()]},
        )
        transaction.commit()

        self.assertEqual(response.status_code, 200)
        self.assertIn("copy_of_doc1", self.portal.objectIds())
        self.assertIn("copy_of_doc2", self.portal.objectIds())

    def test_copy_single_object_no_permissions_raises_403(self):
        self.api_session.auth = ("memberuser", "secret")
        response = self.api_session.post(
            "/@copy", json={"source": self.doc1.absolute_url()}
        )

        self.assertEqual(response.status_code, 403)

    def test_copy_single_object_no_auth_raises_401(self):
        self.api_session.auth = ("nonexistent", "secret")
        response = self.api_session.post(
            "/@copy", json={"source": self.doc1.absolute_url()}
        )

        self.assertEqual(response.status_code, 401)

    def test_move_single_object_no_permissions_raises_403(self):
        self.api_session.auth = ("memberuser", "secret")
        response = self.api_session.post(
            "/@move", json={"source": self.doc1.absolute_url()}
        )

        self.assertEqual(response.status_code, 403)

    def test_move_single_object_no_auth_raises_401(self):
        self.api_session.auth = ("nonexistent", "secret")
        response = self.api_session.post(
            "/@move", json={"source": self.doc1.absolute_url()}
        )

        self.assertEqual(response.status_code, 401)

    def test_move_single_object_no_permission_delete_source_raises_403(self):
        api.user.grant_roles(username="memberuser", obj=self.folder1, roles=["Manager"])
        api.content.transition(obj=self.doc1, transition="publish")
        transaction.commit()

        self.api_session.auth = ("memberuser", "secret")
        response = self.api_session.post(
            "/folder1/@move", json={"source": self.doc1.absolute_url()}
        )

        self.assertEqual(response.status_code, 403)



from DateTime import DateTime
from pkg_resources import get_distribution
from pkg_resources import parse_version
from plone.dexterity.utils import createContentInContainer
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.uuid.interfaces import IMutableUUID
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter

import unittest


HAS_PLONE_6 = parse_version(
    get_distribution("Products.CMFPlone").version
) >= parse_version("6.0.0a1")


class TestCatalogSerializers(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.portal.REQUEST
        self.catalog = getToolByName(self.portal, "portal_catalog")

        # /plone/my-folder
        self.folder = createContentInContainer(self.portal, "Folder", title="My Folder")

        # /plone/my-folder/my-document
        self.doc = createContentInContainer(
            self.folder,
            "Document",
            creation_date=DateTime(2015, 12, 31, 23, 45),
            title="My Document",
        )

        IMutableUUID(self.doc).set("77779ffa110e45afb1ba502f75f77777")
        self.doc.reindexObject()

    def test_lazy_cat_serialization_empty_resultset(self):
        # Force an empty resultset (Products.ZCatalog.Lazy.LazyCat)
        lazy_cat = self.catalog(path="doesnt-exist")
        results = getMultiAdapter((lazy_cat, self.request), ISerializeToJson)()

        self.assertDictEqual(
            {"@id": "http://nohost", "items": [], "items_total": 0}, results
        )

    @unittest.skipUnless(HAS_PLONE_6, "Since Plone 6 the Plone site is indexed ...")
    def test_lazy_map_serialization(self):
        # Test serialization of a Products.ZCatalog.Lazy.LazyMap
        lazy_map = self.catalog()
        results = getMultiAdapter((lazy_map, self.request), ISerializeToJson)()

        self.assertDictContainsSubset({"@id": "http://nohost"}, results)
        self.assertDictContainsSubset({"items_total": 3}, results)
        self.assertEqual(3, len(results["items"]))

    @unittest.skipIf(HAS_PLONE_6, "... before it was not")
    def test_lazy_map_serialization_plone5(self):
        # Test serialization of a Products.ZCatalog.Lazy.LazyMap
        lazy_map = self.catalog()
        results = getMultiAdapter((lazy_map, self.request), ISerializeToJson)()

        self.assertDictContainsSubset({"@id": "http://nohost"}, results)
        self.assertDictContainsSubset({"items_total": 2}, results)
        self.assertEqual(2, len(results["items"]))

    def test_lazy_map_serialization_with_fullobjects(self):
        # Test serialization of a Products.ZCatalog.Lazy.LazyMap
        lazy_map = self.catalog(path="/plone/my-folder/my-document")
        results = getMultiAdapter((lazy_map, self.request), ISerializeToJson)(
            fullobjects=True
        )

        self.assertDictContainsSubset({"@id": "http://nohost"}, results)
        self.assertDictContainsSubset({"items_total": 1}, results)
        self.assertEqual(1, len(results["items"]))
        result_item = results["items"][0]

        self.assertDictContainsSubset(
            {
                "@id": "http://nohost/plone/my-folder/my-document",
                "@type": "Document",
                "changeNote": "",
                "contributors": [],
                "creators": ["test_user_1_"],
                "description": "",
                "effective": None,
                "exclude_from_nav": False,
                "expires": None,
                "id": "my-document",
                "is_folderish": False,
                "language": "",
                "layout": "document_view",
                "parent": {
                    "@id": "http://nohost/plone/my-folder",
                    "@type": "Folder",
                    "description": "",
                    "review_state": "private",
                    "title": "My Folder",
                },
                "relatedItems": [],
                "review_state": "private",
                "rights": "",
                "subjects": [],
                "table_of_contents": None,
                "text": None,
                "title": "My Document",
                "version": "current",
            },
            result_item,
        )

    def test_brain_summary_representation(self):
        lazy_map = self.catalog(path="/plone/my-folder/my-document")
        brain = lazy_map[0]
        result = getMultiAdapter((brain, self.request), ISerializeToJsonSummary)()
        self.assertEqual(
            {
                "@id": "http://nohost/plone/my-folder/my-document",
                "@type": "Document",
                "title": "My Document",
                "description": "",
                "review_state": "private",
            },
            result,
        )



from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from zExceptions import NotFound
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class GroupsDelete(Service):
    """Deletes a user."""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Consume any path segments after /@groups as parameters
        self.params.append(name)
        return self

    @property
    def _get_group_id(self):
        if len(self.params) != 1:
            raise Exception("Must supply exactly one parameter (group id)")
        return self.params[0]

    def _get_group(self, group_id):
        portal = getSite()
        portal_groups = getToolByName(portal, "portal_groups")
        return portal_groups.getGroupById(group_id)

    def reply(self):

        portal_groups = getToolByName(self.context, "portal_groups")
        group = self._get_group(self._get_group_id)

        if not group:
            raise NotFound("Trying to delete a non-existing group.")

        delete_successful = portal_groups.removeGroup(self._get_group_id)
        if delete_successful:
            return self.reply_no_content()
        else:
            return self.reply_no_content(status=404)



from Acquisition import aq_parent
from plone.uuid.interfaces import IUUID
from plone.uuid.interfaces import IUUIDAware
from zope.component import getMultiAdapter


def path2uid(context, link):
    # unrestrictedTraverse requires a string on py3. see:
    # https://github.com/zopefoundation/Zope/issues/674
    if not link:
        return ""
    portal = getMultiAdapter(
        (context, context.REQUEST), name="plone_portal_state"
    ).portal()
    portal_url = portal.portal_url()
    portal_path = "/".join(portal.getPhysicalPath())
    path = link
    context_url = context.absolute_url()
    relative_up = len(context_url.split("/")) - len(portal_url.split("/"))
    if path.startswith(portal_url):
        path = path[len(portal_url) + 1 :]
    if not path.startswith(portal_path):
        path = "{portal_path}/{path}".format(
            portal_path=portal_path, path=path.lstrip("/")
        )
    obj = portal.unrestrictedTraverse(path, None)
    if obj is None or obj == portal:
        return link
    segments = path.split("/")
    suffix = ""
    while not IUUIDAware.providedBy(obj):
        obj = aq_parent(obj)
        suffix += "/" + segments.pop()
    # check if obj is wrong because of acquisition
    if "/".join(obj.getPhysicalPath()) != "/".join(segments):
        return link
    href = relative_up * "../" + "resolveuid/" + IUUID(obj)
    if suffix:
        href += suffix
    return href



from Products.CMFCore.utils import getToolByName
from plone.dexterity.interfaces import IDexterityFTI
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.controlpanels.interfaces import IDexterityTypesControlpanel
from plone.restapi.serializer.controlpanels import SERVICE_ID
from plone.restapi.serializer.controlpanels import ControlpanelSerializeToJson
from plone.restapi.serializer.controlpanels import get_jsonschema_for_controlpanel
from plone.restapi.serializer.converters import json_compatible
from zope.component import adapter
from zope.component import getAllUtilitiesRegisteredFor
from zope.component import queryMultiAdapter
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.i18n import translate


@implementer(ISerializeToJson)
@adapter(IDexterityTypesControlpanel)
class DexterityTypesControlpanelSerializeToJson(ControlpanelSerializeToJson):
    def count(self, portal_type):
        catalog = getToolByName(self.controlpanel.context, "portal_catalog")
        lengths = dict(catalog.Indexes["portal_type"].uniqueValues(withLengths=True))
        return lengths.get(portal_type, 0)

    def serialize_item(self, proxy):
        json_data = {}
        json_schema = {}
        fti = proxy.fti

        overview = queryMultiAdapter(
            (proxy, self.controlpanel.request), name="overview"
        )
        form = overview.form_instance
        json_schema = get_jsonschema_for_controlpanel(
            self.controlpanel,
            self.controlpanel.context,
            self.controlpanel.request,
            form,
        )

        for name, item in form.fields.items():
            serializer = queryMultiAdapter(
                (item.field, fti, self.controlpanel.request), IFieldSerializer
            )
            if serializer:
                value = serializer()
            else:
                value = getattr(fti, name, None)
            json_data[json_compatible(name)] = value

        behaviors = queryMultiAdapter(
            (proxy, self.controlpanel.request), name="behaviors"
        )
        form = behaviors.form_instance
        behaviors_schema = get_jsonschema_for_controlpanel(
            self.controlpanel,
            self.controlpanel.context,
            self.controlpanel.request,
            form,
        )

        behaviors_schema["fieldsets"][0]["id"] = "behaviors"
        behaviors_schema["fieldsets"][0]["title"] = translate(
            "Behaviors", domain="plone", context=self.controlpanel.request
        )
        json_schema["fieldsets"].extend(behaviors_schema["fieldsets"])
        json_schema["properties"].update(behaviors_schema["properties"])

        for name, item in form.fields.items():
            behaviors = getattr(fti, "behaviors", [])
            json_data[json_compatible(name)] = name in behaviors

        # JSON schema
        return {
            "@id": "{}/{}/{}/{}".format(
                self.controlpanel.context.absolute_url(),
                SERVICE_ID,
                self.controlpanel.__name__,
                proxy.__name__,
            ),
            "title": fti.Title(),
            "description": fti.Description(),
            "group": self.controlpanel.group,
            "schema": json_schema,
            "data": json_data,
            "items": [],
        }

    def __call__(self, item=None):
        if item is not None:
            return self.serialize_item(item)

        json = super().__call__()
        json["items"] = []

        portal = getSite()
        portal_url = portal.absolute_url()

        ftis = getAllUtilitiesRegisteredFor(IDexterityFTI)
        for fti in ftis:
            name = fti.__name__
            json["items"].append(
                {
                    "@id": "{}/@controlpanels/dexterity-types/{}".format(
                        portal_url, name
                    ),
                    "@type": name,
                    "meta_type": fti.meta_type,
                    "id": name,
                    "title": fti.Title(),
                    "description": fti.Description(),
                    "count": self.count(name),
                }
            )
        return json



from importlib import import_module
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.schema import SCHEMA_CACHE
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.restapi.testing import RelativeSession
from zope.component import getMultiAdapter
from zope.component import queryUtility

import json
import unittest

HAS_PLONE_6 = getattr(
    import_module("Products.CMFPlone.factory"), "PLONE60MARKER", False
)


class TestSiteSerializer(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.portal.invokeFactory(
            "Document",
            id="doc1",
        )

        fti = queryUtility(IDexterityFTI, name="Plone Site")
        if fti is not None:
            behavior_list = [a for a in fti.behaviors]
            behavior_list.append("volto.blocks")
            fti.behaviors = tuple(behavior_list)
            # Invalidating the cache is required for the FTI to be applied
            # on the existing object
            SCHEMA_CACHE.invalidate("Plone Site")

    def serialize(self):
        serializer = getMultiAdapter((self.portal, self.request), ISerializeToJson)
        return serializer()

    def test_serializer_returns_json_serializeable_object(self):
        obj = self.serialize()
        self.assertTrue(isinstance(json.dumps(obj), str), "Not JSON serializable")

    def test_serializer_includes_title(self):
        obj = self.serialize()
        self.assertIn("title", obj)
        self.assertEqual("Plone site", obj["title"])

    def test_get_is_folderish(self):
        obj = self.serialize()
        self.assertIn("is_folderish", obj)
        self.assertEqual(True, obj["is_folderish"])

    @unittest.skipIf(
        HAS_PLONE_6,
        "This test is only intended to run for Plone 5 and the blocks behavior site root hack enabled",
    )
    def test_resolveuids_get_serialized_in_serializer_plone5(self):
        blocks = {
            "0358abe2-b4f1-463d-a279-a63ea80daf19": {
                "@type": "foo",
                "url": f"resolveuid/{self.portal.doc1.UID()}",
            },
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
        }
        self.portal.blocks = json.dumps(blocks)
        obj = self.serialize()
        self.assertEqual(
            obj["blocks"]["0358abe2-b4f1-463d-a279-a63ea80daf19"]["url"],
            self.portal.doc1.absolute_url(),
        )

    @unittest.skipIf(
        not HAS_PLONE_6,
        "This test is only intended to run for Plone 6 and DX site root enabled",
    )
    def test_resolveuids_get_serialized_in_serializer(self):
        blocks = {
            "0358abe2-b4f1-463d-a279-a63ea80daf19": {
                "@type": "foo",
                "url": f"resolveuid/{self.portal.doc1.UID()}",
            },
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
        }
        self.portal.blocks = blocks
        obj = self.serialize()
        self.assertEqual(
            obj["blocks"]["0358abe2-b4f1-463d-a279-a63ea80daf19"]["url"],
            self.portal.doc1.absolute_url(),
        )


class TestSiteSerializationFunctional(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(f"{self.portal_url}/++api++", test=self)
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

    def tearDown(self):
        self.api_session.close()

    def test_site_root_get_request(self):
        response = self.api_session.get("")

        self.assertEqual(response.json()["@id"], self.portal.absolute_url())



from pkg_resources import get_distribution
from pkg_resources import parse_version
from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services import Service
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import getMultiAdapter

zcatalog_version = get_distribution("Products.ZCatalog").version
if parse_version(zcatalog_version) >= parse_version("5.1"):
    SUPPORT_NOT_UUID_QUERIES = True
else:
    SUPPORT_NOT_UUID_QUERIES = False


class QuerystringSearchPost(Service):
    """Returns the querystring search results given a p.a.querystring data."""

    def reply(self):
        data = json_body(self.request)
        query = data.get("query", None)
        b_start = int(data.get("b_start", 0))
        b_size = int(data.get("b_size", 25))
        sort_on = data.get("sort_on", None)
        sort_order = data.get("sort_order", None)
        limit = int(data.get("limit", 1000))
        fullobjects = data.get("fullobjects", False)

        if query is None:
            raise Exception("No query supplied")

        if sort_order:
            sort_order = "descending" if sort_order == "descending" else "ascending"

        querybuilder = getMultiAdapter(
            (self.context, self.request), name="querybuilderresults"
        )

        querybuilder_parameters = dict(
            query=query,
            brains=True,
            b_start=b_start,
            b_size=b_size,
            sort_on=sort_on,
            sort_order=sort_order,
            limit=limit,
        )

        # Exclude "self" content item from the results when ZCatalog supports NOT UUID
        # queries and it is called on a content object.
        if not IPloneSiteRoot.providedBy(self.context) and SUPPORT_NOT_UUID_QUERIES:
            querybuilder_parameters.update(
                dict(custom_query={"UID": {"not": self.context.UID()}})
            )

        results = querybuilder(**querybuilder_parameters)

        results = getMultiAdapter((results, self.request), ISerializeToJson)(
            fullobjects=fullobjects
        )
        return results



from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from zExceptions import BadRequest
from zope.component import queryMultiAdapter
from zope.component.hooks import getSite
from zope.interface import alsoProvides

import plone.protect.interfaces


class GroupsPost(Service):
    """Creates a new group."""

    def reply(self):
        portal = getSite()
        data = json_body(self.request)

        groupname = data.get("groupname", None)

        if not groupname:
            raise BadRequest("Property 'groupname' is required")

        email = data.get("email", None)
        title = data.get("title", None)
        description = data.get("description", None)
        roles = data.get("roles", None)
        groups = data.get("groups", None)
        users = data.get("users", [])

        properties = {"title": title, "description": description, "email": email}

        gtool = getToolByName(self.context, "portal_groups")
        regtool = getToolByName(self.context, "portal_registration")

        if not regtool.isMemberIdAllowed(groupname):
            raise BadRequest("The group name you entered is not valid.")

        already_exists = gtool.getGroupById(groupname)
        if already_exists:
            raise BadRequest("The group name you entered already exists.")

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        success = gtool.addGroup(
            groupname,
            roles,
            groups,
            properties=properties,
            title=title,
            description=description,
        )
        if not success:
            raise BadRequest(f"Error occurred, could not add group {groupname}.")

        # Add members
        group = gtool.getGroupById(groupname)
        for userid in users:
            group.addMember(userid)

        self.request.response.setStatus(201)
        self.request.response.setHeader(
            "Location", portal.absolute_url() + "/@groups/" + groupname
        )
        serializer = queryMultiAdapter((group, self.request), ISerializeToJson)
        return serializer()



from DateTime import DateTime
from datetime import datetime
from plone.app.discussion.comment import Comment
from plone.app.layout.viewlets.content import ContentHistoryViewlet
from plone.dexterity.content import DexterityContent
from plone.locking.lockable import TTWLockable
from Products.CMFCore.WorkflowTool import _marker
from Products.CMFCore.WorkflowTool import WorkflowTool
from plone.restapi.serializer.working_copy import WorkingCopyInfo


_originals = {
    "WorkflowTool.getInfoFor": WorkflowTool.getInfoFor,
    "ContentHistoryViewlet.fullHistory": ContentHistoryViewlet.fullHistory,
    "TTWLockable.lock_info": TTWLockable.lock_info,
    "WorkingCopyInfo.created": WorkingCopyInfo.created,
}


class StaticTime:
    """ContextManager to patch accessor methods that return dynamic timestamps,
    like creation and modification dates, with ones that return static
    timestamps.

    This is needed during testing to get stable serialization output,
    especially during the tests that dump HTTP response examples to the
    filesystem, because the content of those responses would otherwise
    change all the time.

    Specifically monkey-patching a few accessors (getter methods) in a
    targeted way like this was chosen as an alternative to freezing time
    entirely using libraries like `freezegun`, because that approach has
    caused issues in the past with code that assumes a monotonic clock, like
    the generation of ZODB transaction IDs.

    Instead, this helper aims to selectively patch only what we currently
    need to get stable test runs and HTTP response examples.

    The bulk of cases is addressed by patching `creation_date` and
    `modification_date` for Dexterity objects to return static times. In
    addition, a couple other places need some attention.

    These places will be patched on start() (and unpatched on stop()):

    - DexterityContent
        - creation_date
        - modification_date

    - p.a.discussion Comment
        - creation_date
        - modification_date

    - WorkflowTool.getInfoFor
        - (if asked for 'review_history')

    - ContentHistoryViewlet
        - fullHistory

    - TTWLockable
        - lock_info

    """

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, _type, exc, _traceback):
        self.stop()

    def __init__(
        self,
        created=datetime(1995, 7, 31, 13, 45),
        modified=datetime(1995, 7, 31, 17, 30),
    ):
        self.static_created = created
        self.static_modified = modified
        """Set up a static time helper.

        If given, the Python datetimes for `created` and `modified` will be
        used to determine the timestamps that should be returned by the
        static getters.

        Ensuring the appropriate type (Python datetime vs. Zope DateTime
        vs. float) is the responsiblity of the patched getter method. The user
        of the StaticTime class shouldn't need to care about it, and be able
        to pass Python datetimes that then will be casted appropriately if
        necessary.

        The `modified` timestamp will also be used as a basis when producing
        a sequence of fake times, for example when patching times for
        successive events in the review history.
        """

    def start(self):
        """Patch the respective getters so that they return static times."""
        # Patch created and modified times for DexterityContent.
        # creation_date and modification_date are instance-level attributes
        # on DX object that get initialized with datetime.now() during
        # __init__. In order to fake the returned times we patch a property
        # onto the class which will shadow these instance attributes.
        DexterityContent.creation_date = property(
            static_creation_date_getter_factory(self.static_created), nop_setter
        )
        DexterityContent.modification_date = property(
            static_modification_date_getter_factory(self.static_modified), nop_setter
        )

        # Patch the lightweight p.a.discussion 'Comment' type. Its dates are
        # Python datetimes, unlike DX Content types which use zope DateTimes.
        Comment.creation_date = property(
            static_creation_date_getter_factory(self.static_created, type_=datetime),
            nop_setter,
        )
        Comment.modification_date = property(
            static_modification_date_getter_factory(
                self.static_modified, type_=datetime
            ),
            nop_setter,
        )

        WorkflowTool.getInfoFor = static_get_info_for_factory(self.static_modified)

        ContentHistoryViewlet.fullHistory = static_full_history_factory(
            self.static_modified
        )

        TTWLockable.lock_info = static_lock_info_factory(self.static_modified)

        WorkingCopyInfo.created = static_wc_info_factory(self.static_created)

    def stop(self):
        """Undo all the patches."""
        TTWLockable.lock_info = _originals["TTWLockable.lock_info"]
        ContentHistoryViewlet.fullHistory = _originals[
            "ContentHistoryViewlet.fullHistory"
        ]
        WorkflowTool.getInfoFor = _originals["WorkflowTool.getInfoFor"]

        Comment.modification_date = None
        Comment.creation_date = None

        WorkingCopyInfo.created = _originals["WorkingCopyInfo.created"]

        del DexterityContent.modification_date
        del DexterityContent.creation_date


def static_get_info_for_factory(dt_value):
    """Returns a static time replacement for WorkflowTool.getInfoFor
    configured with the given datetime value as a base.
    """
    if isinstance(dt_value, datetime):
        dt_value = DateTime(dt_value)

    def static_get_info_for(self, ob, name, default=_marker, wf_id=None, *args, **kw):
        """This replacement function will, if 'review_history' is requested,
        replace timestamps in the returned list of dicts with static times.

        The timestamps for the successive events in this list will use the
        dt_value (which defaults to modification date) as a base, and then
        move forward in hourly steps to retain the chronological sequence of
        events.

        In other words, they will be stable (static), but different for each
        event, and should still reflect proper order of events.
        """
        res = _originals["WorkflowTool.getInfoFor"](
            self, ob, name, default=default, wf_id=wf_id, *args, **kw
        )
        if name == "review_history":
            base_date = dt_value

            # The ContentHistoryViewlet.fullHistory method assembles results
            # from both the review_history (i.e., this method's result) and
            # the revision history, and intertwines their elements in
            # chronological order. That doesn't work if we already return
            # faked timestamps here in that case.
            #
            # We therefore need to recognize this case by inspecting the call
            # stack to check whether we got called by the fullHistory method,
            # and in that case return the original results. (Since fullHistory
            # will also be patched, timestamps in the combined results will be
            # replaced there.)
            import traceback

            stack = traceback.format_stack()
            if "static_full_history" in str(stack):
                return res

            for idx, item in enumerate(res):
                fake_date = base_date + (idx / 24.0)  # plus one hour
                if "time" in item:
                    item["time"] = fake_date

        return res

    return static_get_info_for


def static_full_history_factory(dt_value):
    """Returns a static time replacement for ContentHistoryViewlet.fullHistory
    configured with the given datetime value as a base.
    """
    if isinstance(dt_value, datetime):
        dt_value = DateTime(dt_value)

    def static_full_history(self):
        """This replacement function will replace timestamps in the returned
        list of dicts with static times.

        The timestamps for the successive events in this list will use the
        dt_value (which defaults to modification date) as a base, and then
        move forward in hourly steps to retain the chronological sequence of
        events.

        In other words, they will be stable (static), but different for each
        event, and should still reflect proper order of events.
        """
        actions = _originals["ContentHistoryViewlet.fullHistory"](self)

        base_date = dt_value
        for idx, action in enumerate(actions):
            if "time" in action:
                fake_date = base_date + (idx / 24.0)  # plus one hour

                # Depending on the kind of action, timestamps may either
                # be zope DateTimes or floats. Let's reserve the same type.
                if isinstance(action["time"], float):
                    action["time"] = float(fake_date)
                elif isinstance(action["time"], DateTime):
                    action["time"] = fake_date
                else:
                    raise Exception("Don't know how to patch %r" % action["time"])

        return actions

    return static_full_history


def static_creation_date_getter_factory(dt_value, type_=DateTime):
    """Returns a static time replacement for creation date accessors,
    configured with the given datetime value and the indicated type_.
    """
    if isinstance(dt_value, datetime) and type_ is DateTime:
        dt_value = DateTime(dt_value)

    elif isinstance(dt_value, DateTime) and type_ is datetime:
        dt_value = dt_value.asdatetime()

    def static_creation_date_getter(self):
        return dt_value

    return static_creation_date_getter


def static_modification_date_getter_factory(dt_value, type_=DateTime):
    """Returns a static time replacement for modification date accessors,
    configured with the given datetime value and the indicated type_.
    """
    if isinstance(dt_value, datetime) and type_ is DateTime:
        dt_value = DateTime(dt_value)

    elif isinstance(dt_value, DateTime) and type_ is datetime:
        dt_value = dt_value.asdatetime()

    def static_modification_date_getter(self, value=None):
        return dt_value

    return static_modification_date_getter


def static_lock_info_factory(dt_value):
    """Returns a static time replacement for TTWLockable.lock_info
    configured with the given datetime value as a base.
    """
    if isinstance(dt_value, datetime):
        dt_value = DateTime(dt_value)

    def static_lock_info(self):
        """This replacement function will replace timestamps in the returned
        list of lock_info dicts with static times.

        The timestamps for the successive lock_infos in this list will use the
        dt_value (which defaults to modification date) as a base, and then
        move forward in hourly steps to retain the chronological sequence of
        events.

        In other words, they will be stable (static), but different for each
        lock_info, and should still reflect proper order of events.
        """
        infos = _originals["TTWLockable.lock_info"](self)
        base_date = dt_value
        for idx, info in enumerate(infos):
            fake_date = base_date + (idx / 24.0)  # plus one hour
            if "time" in info:
                info["time"] = float(fake_date)

        return infos

    return static_lock_info


def static_wc_info_factory(dt_value, type_=DateTime):
    """Returns a static time replacement for creation date accessors,
    configured with the given datetime value and the indicated type_ for
    working copy support.
    """
    if isinstance(dt_value, datetime) and type_ is DateTime:
        dt_value = DateTime(dt_value)

    elif isinstance(dt_value, DateTime) and type_ is datetime:
        dt_value = dt_value.asdatetime()

    def static_wc_info(self):
        return dt_value

    return static_wc_info


def nop_setter(self, value):
    pass



from plone.restapi.search.utils import unflatten_dotted_dict

import unittest


class TestUnflattenDottedDict(unittest.TestCase):
    def test_unflattens_dotted_dict(self):
        dct = {"a.b.X": 1, "a.b.Y": 2, "a.foo": 3, "bar": 4}
        self.assertEqual(
            {"a": {"b": {"X": 1, "Y": 2}, "foo": 3}, "bar": 4},
            unflatten_dotted_dict(dct),
        )

    def test_works_on_empty_dict(self):
        self.assertEqual({}, unflatten_dotted_dict({}))

    def test_works_with_list_values(self):
        dct = {"path.query": ["foo", "bar"], "path.depth": 2}
        self.assertEqual(
            {"path": {"query": ["foo", "bar"], "depth": 2}}, unflatten_dotted_dict(dct)
        )

    def test_leaves_regular_keys_untouched(self):
        dct = {"foo": 1, "bar": 2}
        self.assertEqual(dct, unflatten_dotted_dict(dct))



from plone.restapi.services import Service
from plone.restapi.services.addons.addons import Addons
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class AddonsGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []
        self.addons = Addons(context, request)
        self.query = self.request.form.copy()

    def publishTraverse(self, request, name):
        # Consume any path segments after /@addons as parameters
        self.params.append(name)
        return self

    def reply(self):
        all_addons = self.addons.get_addons()

        if self.params:
            if self.params[0] in all_addons:
                return self.addons.serializeAddon(all_addons[self.params[0]])
            else:
                return []

        result = {
            "items": {"@id": f"{self.context.absolute_url()}/@addons"},
        }
        addons_data = []
        for addon in all_addons.values():
            addons_data.append(self.addons.serializeAddon(addon))

        if len(self.query) > 0 and len(self.params) == 0:
            upgradeables = self.query.get("upgradeable", "")
            if upgradeables:
                addons_data = [
                    addon
                    for addon in addons_data
                    if addon.get("upgrade_info", False)
                    and addon["upgrade_info"].get("available", False)
                ]

        result["items"] = addons_data
        self.request.response.setStatus(200)
        return result



from plone.app.contenttypes.interfaces import ICollection
from plone.restapi.batching import HypermediaBatch
from plone.restapi.deserializer import boolean_value
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.serializer.dxcontent import SerializeToJson
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.interface import Interface


@implementer(ISerializeToJson)
@adapter(ICollection, Interface)
class SerializeCollectionToJson(SerializeToJson):
    def __call__(self, version=None, include_items=True):
        result = super().__call__(version=version)

        include_items = self.request.form.get("include_items", include_items)
        include_items = boolean_value(include_items)
        if include_items:
            results = self.context.results(batch=False)
            batch = HypermediaBatch(self.request, results)

            if not self.request.form.get("fullobjects"):
                result["@id"] = batch.canonical_url
            result["items_total"] = batch.items_total
            if batch.links:
                result["batching"] = batch.links

            if "fullobjects" in list(self.request.form):
                result["items"] = [
                    getMultiAdapter(
                        (brain.getObject(), self.request), ISerializeToJson
                    )()
                    for brain in batch
                ]
            else:
                result["items"] = [
                    getMultiAdapter((brain, self.request), ISerializeToJsonSummary)()
                    for brain in batch
                ]

        return result



from DateTime import DateTime
from plone.app.contentlisting.interfaces import IContentListingObject
from plone.app.testing import popGlobalRegistry
from plone.app.testing import pushGlobalRegistry
from plone.dexterity.utils import createContentInContainer
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.restapi.testing import register_static_uuid_utility
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter
from zope.component.hooks import getSite

import Missing
import unittest


class TestSummarySerializers(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        pushGlobalRegistry(getSite())
        register_static_uuid_utility(prefix="c6dcbd55ab2746e199cd4ed458")

        self.catalog = getToolByName(self.portal, "portal_catalog")

        self.doc1 = createContentInContainer(
            self.portal,
            "DXTestDocument",
            id="doc1",
            title="Lorem Ipsum",
            description="Description",
        )

        self.doc1.creation_date = DateTime("2016-01-21T01:14:48+00:00")
        self.doc1.modification_date = DateTime("2017-01-21T01:14:48+00:00")
        self.doc1.reindexObject(["modified"])

    def tearDown(self):
        popGlobalRegistry(getSite())

    def test_site_root_summary(self):
        summary = getMultiAdapter(
            (self.portal, self.request), ISerializeToJsonSummary
        )()

        self.assertDictEqual(
            {
                "@id": "http://nohost/plone",
                "@type": "Plone Site",
                "title": "Plone site",
                "description": "",
            },
            summary,
        )

    def test_brain_summary(self):
        brain = self.catalog(UID=self.doc1.UID())[0]
        summary = getMultiAdapter((brain, self.request), ISerializeToJsonSummary)()

        self.assertDictEqual(
            {
                "@id": "http://nohost/plone/doc1",
                "@type": "DXTestDocument",
                "title": "Lorem Ipsum",
                "description": "Description",
                "review_state": "private",
            },
            summary,
        )

        # Must also work if we're dealing with a CatalogContentListingObject
        # (because the brain has already been adapted to IContentListingObject,
        # as is the case for collection results)
        listing_obj = IContentListingObject(brain)
        summary = getMultiAdapter(
            (listing_obj, self.request), ISerializeToJsonSummary
        )()

        self.assertDictEqual(
            {
                "@id": "http://nohost/plone/doc1",
                "@type": "DXTestDocument",
                "title": "Lorem Ipsum",
                "description": "Description",
                "review_state": "private",
            },
            summary,
        )

    def test_brain_summary_with_missing_value(self):
        brain = self.catalog(UID=self.doc1.UID())[0]
        brain.Description = Missing.Value

        summary = getMultiAdapter((brain, self.request), ISerializeToJsonSummary)()

        self.assertDictEqual(
            {
                "@id": "http://nohost/plone/doc1",
                "@type": "DXTestDocument",
                "title": "Lorem Ipsum",
                "description": None,
                "review_state": "private",
            },
            summary,
        )

    def test_brain_summary_includes_additional_metadata_fields(self):
        brain = self.catalog(UID=self.doc1.UID())[0]
        self.request.form.update({"metadata_fields": ["UID", "Creator"]})
        summary = getMultiAdapter((brain, self.request), ISerializeToJsonSummary)()

        self.assertDictEqual(
            {
                "@id": "http://nohost/plone/doc1",
                "@type": "DXTestDocument",
                "UID": "c6dcbd55ab2746e199cd4ed458000001",
                "Creator": "test_user_1_",
                "title": "Lorem Ipsum",
                "description": "Description",
                "review_state": "private",
            },
            summary,
        )

    def test_brain_summary_includes_all_metadata_fields(self):
        brain = self.catalog(UID=self.doc1.UID())[0]
        self.request.form.update({"metadata_fields": "_all"})
        summary = getMultiAdapter((brain, self.request), ISerializeToJsonSummary)()

        # mime_type was added in Plone 5.1
        # Make sure tests pass on older Plone versions
        if "mime_type" not in summary:
            summary["mime_type"] = "text/plain"

        self.maxDiff = None
        self.assertLessEqual(
            {
                "@id": "http://nohost/plone/doc1",
                "@type": "DXTestDocument",
                "CreationDate": "2016-01-21T01:14:48+00:00",
                "Creator": "test_user_1_",
                "Date": "2017-01-21T01:14:48+00:00",
                "Description": "Description",
                "EffectiveDate": "None",
                "ExpirationDate": "None",
                "ModificationDate": "2017-01-21T01:14:48+00:00",
                "Subject": [],
                "Title": "Lorem Ipsum",
                "Type": "DX Test Document",
                "UID": "c6dcbd55ab2746e199cd4ed458000001",
                "author_name": None,
                "cmf_uid": None,
                "commentators": [],
                "created": "2016-01-21T01:14:48+00:00",
                "description": "Description",
                "effective": "1969-12-31T00:00:00+00:00",
                "end": None,
                "exclude_from_nav": False,
                "expires": "2499-12-31T00:00:00+00:00",
                "getIcon": None,
                "getId": "doc1",
                "getObjSize": "0 KB",
                "getPath": "/plone/doc1",
                "getRemoteUrl": None,
                "getURL": "http://nohost/plone/doc1",
                "id": "doc1",
                "in_response_to": None,
                "is_folderish": False,
                "last_comment_date": None,
                "listCreators": ["test_user_1_"],
                "location": None,
                "mime_type": "text/plain",
                "modified": "2017-01-21T01:14:48+00:00",
                "portal_type": "DXTestDocument",
                "review_state": "private",
                "start": None,
                "sync_uid": None,
                "title": "Lorem Ipsum",
                "total_comments": 0,
            }.items(),
            summary.items(),
        )

    def test_dx_type_summary(self):
        summary = getMultiAdapter((self.doc1, self.request), ISerializeToJsonSummary)()

        self.assertDictEqual(
            {
                "@id": "http://nohost/plone/doc1",
                "@type": "DXTestDocument",
                "title": "Lorem Ipsum",
                "description": "Description",
                "review_state": "private",
            },
            summary,
        )



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.dexterity.utils import createContentInContainer
from plone.registry.interfaces import IRegistry
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from Products.CMFPlone.interfaces.controlpanel import INavigationSchema
from zope.component import getUtility

import transaction
import unittest


class TestServicesNavigation(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.folder = createContentInContainer(
            self.portal, "Folder", id="folder", title="Some Folder"
        )
        self.folder2 = createContentInContainer(
            self.portal, "Folder", id="folder2", title="Some Folder 2"
        )
        self.subfolder1 = createContentInContainer(
            self.folder, "Folder", id="subfolder1", title="SubFolder 1"
        )
        self.subfolder2 = createContentInContainer(
            self.folder, "Folder", id="subfolder2", title="SubFolder 2"
        )
        self.thirdlevelfolder = createContentInContainer(
            self.subfolder1,
            "Folder",
            id="thirdlevelfolder",
            title="Third Level Folder",
        )
        self.fourthlevelfolder = createContentInContainer(
            self.thirdlevelfolder,
            "Folder",
            id="fourthlevelfolder",
            title="Fourth Level Folder",
        )
        createContentInContainer(self.folder, "Document", id="doc1", title="A document")
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_navigation_service(self):
        response = self.api_session.get(
            "/folder/@navigation", params={"expand.navigation.depth": 2}
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["items"]), 3)
        self.assertEqual(response.json()["items"][1]["title"], "Some Folder")
        self.assertEqual(len(response.json()["items"][1]["items"]), 3)
        self.assertEqual(len(response.json()["items"][2]["items"]), 0)

        response = self.api_session.get(
            "/folder/@navigation", params={"expand.navigation.depth": 3}
        )

        self.assertEqual(len(response.json()["items"][1]["items"][0]["items"]), 1)
        self.assertEqual(
            response.json()["items"][1]["items"][0]["items"][0]["title"],
            "Third Level Folder",
        )
        self.assertEqual(
            len(response.json()["items"][1]["items"][0]["items"][0]["items"]),
            0,
        )

        response = self.api_session.get(
            "/folder/@navigation", params={"expand.navigation.depth": 4}
        )

        self.assertEqual(
            len(response.json()["items"][1]["items"][0]["items"][0]["items"]),
            1,
        )
        self.assertEqual(
            response.json()["items"][1]["items"][0]["items"][0]["items"][0][
                "title"
            ],  # noqa
            "Fourth Level Folder",
        )

    def test_dont_broke_with_contents_without_review_state(self):
        createContentInContainer(
            self.portal,
            "File",
            id="example-file",
            title="Example file",
        )
        createContentInContainer(
            self.folder,
            "File",
            id="example-file-1",
            title="Example file 1",
        )
        transaction.commit()

        response = self.api_session.get("/folder/@navigation")
        self.assertIsNone(response.json()["items"][3]["review_state"])

        response = self.api_session.get(
            "/folder/@navigation", params={"expand.navigation.depth": 2}
        )
        self.assertIsNone(response.json()["items"][1]["items"][3]["review_state"])

    def test_navigation_sorting(self):
        createContentInContainer(
            self.portal,
            "File",
            id="example-file",
            title="Example file",
        )
        createContentInContainer(
            self.folder,
            "File",
            id="example-file-1",
            title="Example file 1",
        )
        transaction.commit()
        response = self.api_session.get(
            "/folder/@navigation", params={"expand.navigation.depth": 2}
        ).json()

        contents = response["items"][1]["items"]
        self.assertEqual(
            [p["@id"].replace(self.portal.absolute_url(), "") for p in contents],
            [
                "/folder/subfolder1",
                "/folder/subfolder2",
                "/folder/doc1",
                "/folder/example-file-1",
            ],
        )

        self.portal["folder"].moveObjectsUp(["example-file-1"])
        transaction.commit()
        response = self.api_session.get(
            "/folder/@navigation", params={"expand.navigation.depth": 2}
        ).json()
        contents = response["items"][1]["items"]
        self.assertEqual(
            [p["@id"].replace(self.portal.absolute_url(), "") for p in contents],
            [
                "/folder/subfolder1",
                "/folder/subfolder2",
                "/folder/example-file-1",
                "/folder/doc1",
            ],
        )

    def test_use_nav_title_when_available_and_set(self):
        registry = getUtility(IRegistry)
        settings = registry.forInterface(INavigationSchema, prefix="plone")
        displayed_types = settings.displayed_types
        settings.displayed_types = tuple(list(displayed_types) + ["DXTestDocument"])

        title = "Example Document"
        nav_title = "Fancy title"

        createContentInContainer(
            self.folder,
            "DXTestDocument",
            id="example-dx-document",
            title=title,
            nav_title=nav_title,
        )
        transaction.commit()

        response = self.api_session.get(
            "/folder/@navigation", params={"expand.navigation.depth": 2}
        )

        self.assertEqual(response.json()["items"][1]["items"][-1]["title"], nav_title)



from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITranslationManager
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.dexterity.utils import createContentInContainer
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from zope.interface import alsoProvides
from plone.app.testing import login


import transaction
import unittest


class TestServicesBreadcrumbs(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.folder = createContentInContainer(
            self.portal, "Folder", id="folder", title="Some Folder"
        )
        createContentInContainer(self.folder, "Document", id="doc1", title="A document")
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_breadcrumbs(self):
        response = self.api_session.get("/folder/doc1/@breadcrumbs")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "@id": self.portal_url + "/folder/doc1/@breadcrumbs",
                "root": self.portal_url,
                "items": [
                    {
                        "@id": self.portal_url + "/folder",
                        "title": "Some Folder",
                    },
                    {
                        "@id": self.portal_url + "/folder/doc1",
                        "title": "A document",
                    },
                ],
            },
        )


class TestServicesMultilingualBreadcrumbs(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        self.request = self.layer["request"]

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        login(self.portal, SITE_OWNER_NAME)
        self.en_content = createContentInContainer(
            self.portal["en"], "Document", title="Test document"
        )
        self.es_content = createContentInContainer(
            self.portal["es"], "Document", title="Test document"
        )
        ITranslationManager(self.en_content).register_translation("es", self.es_content)
        self.folder = createContentInContainer(
            self.portal["es"], "Folder", id="folder", title="Some Folder"
        )
        createContentInContainer(self.folder, "Document", id="doc1", title="A document")
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_breadcrumbs_multilingual(self):
        response = self.api_session.get("/es/folder/doc1/@breadcrumbs")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "@id": self.portal_url + "/es/folder/doc1/@breadcrumbs",
                "root": self.portal_url + "/es",
                "items": [
                    {
                        "@id": self.portal_url + "/es/folder",
                        "title": "Some Folder",
                    },
                    {
                        "@id": self.portal_url + "/es/folder/doc1",
                        "title": "A document",
                    },
                ],
            },
        )



from plone.app.discussion.interfaces import IComment
from plone.app.discussion.interfaces import IConversation
from plone.restapi.batching import HypermediaBatch
from plone.restapi.interfaces import IJsonCompatible
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services.discussion.utils import can_delete
from plone.restapi.services.discussion.utils import can_delete_own
from plone.restapi.services.discussion.utils import can_edit
from plone.restapi.services.discussion.utils import delete_own_comment_allowed
from plone.restapi.services.discussion.utils import edit_comment_allowed
from Products.CMFCore.utils import getToolByName
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.publisher.interfaces import IRequest


@implementer(ISerializeToJson)
@adapter(IConversation, IRequest)
class ConversationSerializer:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        # We'll batch the threads
        results = list(self.context.getThreads())
        batch = HypermediaBatch(self.request, results)

        results = {}
        results["@id"] = batch.canonical_url

        results["items_total"] = batch.items_total
        if batch.links:
            results["batching"] = batch.links

        results["items"] = [
            getMultiAdapter((thread["comment"], self.request), ISerializeToJson)()
            for thread in batch
        ]

        return results


@implementer(ISerializeToJson)
@adapter(IComment, IRequest)
class CommentSerializer:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, include_items=True):
        content_url = self.context.__parent__.__parent__.absolute_url()
        comments_url = f"{content_url}/@comments"
        url = f"{comments_url}/{self.context.id}"

        if self.context.in_reply_to:
            parent_url = f"{comments_url}/{self.context.in_reply_to}"
            in_reply_to = str(self.context.in_reply_to)
        else:
            parent_url = None
            in_reply_to = None

        doc_allowed = delete_own_comment_allowed()
        delete_own = doc_allowed and can_delete_own(self.context)

        if self.context.mime_type == "text/plain":
            text_data = self.context.text
            text_mime_type = self.context.mime_type
        else:
            text_data = self.context.getText()
            text_mime_type = "text/html"
        return {
            "@id": url,
            "@type": self.context.portal_type,
            "@parent": parent_url,
            "comment_id": str(self.context.id),
            "in_reply_to": in_reply_to,
            "text": {"data": text_data, "mime-type": text_mime_type},
            "user_notification": self.context.user_notification,
            "author_username": self.context.author_username,
            "author_name": self.context.author_name,
            "author_image": self.get_author_image(self.context.author_username),
            "creation_date": IJsonCompatible(self.context.creation_date),
            "modification_date": IJsonCompatible(
                self.context.modification_date
            ),  # noqa
            "is_editable": edit_comment_allowed() and can_edit(self.context),
            "is_deletable": can_delete(self.context) or delete_own,
        }

    def get_author_image(self, username=None):
        if username is None:
            return
        portal_membership = getToolByName(self.context, "portal_membership", None)
        image = portal_membership.getPersonalPortrait(username).absolute_url()
        if image.endswith("defaultUser.png"):
            return
        return image



from plone.restapi.interfaces import IExpandableElement
from zope.component import getAdapters

import warnings


def expandable_elements(context, request):
    """Returns a dict containing all expandable elements."""
    expands = request.form.get("expand", [])
    if isinstance(expands, str):
        if "," in expands:
            # deprecated use of expands (with commas)
            warnings.warn(
                "``expand`` as a string parameter separated with commas is deprecated and will be removed in plone.restapi 9.0. Use ``expand:list`` instead.",
                DeprecationWarning,
            )
            expands = expands.split(",")
        else:
            # allow still the use of non marked :list parameters present
            expands = [expands]

    elements = getAdapters((context, request), IExpandableElement)
    res = {}
    for element in elements:
        if element[0] in expands:
            update_dict_recursively(res, element[1](expand=True))
        else:
            update_dict_recursively(res, element[1](expand=False))
    return {"@components": res}


def update_dict_recursively(d, u):
    for key, value in u.items():
        if isinstance(value, dict):
            r = update_dict_recursively(d.get(key, {}), value)
            d[key] = r
        else:
            d[key] = u[key]
    return d



""" Locking
"""
from plone import api
from datetime import datetime
from plone.locking.interfaces import ILockable


def creator_name(username):
    user = api.user.get(username=username)
    return user.getProperty("fullname") or username


def creator_url(username):
    url = api.portal.get().absolute_url()
    return f"{url}/author/{username}"


def creation_date(timestamp):
    return datetime.fromtimestamp(timestamp).isoformat()


def lock_info(obj):
    """Returns lock information about the given object."""
    lockable = ILockable(obj, None)
    if lockable is not None:
        info = {"locked": lockable.locked(), "stealable": lockable.stealable()}
        lock_info = lockable.lock_info()
        if len(lock_info) > 0:
            creator = lock_info[0]["creator"]
            info["creator"] = creator
            info["creator_name"] = creator_name(creator)
            info["creator_url"] = creator_url(creator)
            created = lock_info[0]["time"]
            info["time"] = created
            info["created"] = creation_date(created)
            info["token"] = lock_info[0]["token"]
            lock_type = lock_info[0]["type"]
            if lock_type:
                info["name"] = lock_info[0]["type"].__name__
            lock_item = webdav_lock(obj)
            if lock_item:
                info["timeout"] = lock_item.getTimeout()
        return info
    return {}


def webdav_lock(obj):
    """Returns the WebDAV LockItem"""
    lockable = ILockable(obj, None)
    if lockable is None:
        return

    lock_info = lockable.lock_info()
    if len(lock_info) > 0:
        token = lock_info[0]["token"]
        return obj.wl_getLock(token)


def is_locked(obj, request):
    """Returns true if the object is locked and the request doesn't contain
    the lock token.
    """
    lockable = ILockable(obj, None)
    if lockable is None:
        return False
    if lockable.locked():
        token = request.getHeader("Lock-Token", "")
        lock_info = lockable.lock_info()
        if len(lock_info) > 0 and lock_info[0]["token"] == token:
            return False
        return True
    return False



from Acquisition import aq_inner
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from zope.i18n import translate


class RolesGet(Service):
    def reply(self):
        pmemb = getToolByName(aq_inner(self.context), "portal_membership")
        roles = [r for r in pmemb.getPortalRoles() if r != "Owner"]
        return [
            {
                "@type": "role",
                "@id": f"{self.context.absolute_url()}/@roles/{r}",
                "id": r,
                "title": translate(r, context=self.request, domain="plone"),
            }
            for r in roles
        ]



from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import IPloneRestapiLayer
from plone.restapi.services import Service
from plone.restapi.types.utils import add_field
from plone.restapi.types.utils import add_fieldset
from zExceptions import BadRequest
from zope.component import queryMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import noLongerProvides
from zope.publisher.interfaces import IPublishTraverse
import plone.protect.interfaces


@implementer(IPublishTraverse)
class TypesPost(Service):
    """Creates a new field/fieldset"""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Treat any path segments after /@types as parameters
        self.params.append(name)
        return self

    def reply(self):
        if not self.params:
            raise BadRequest("Missing parameter typename")

        data = json_body(self.request)

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        # Make sure we get the right dexterity-types adapter
        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        name = self.params.pop()
        context = queryMultiAdapter(
            (self.context, self.request), name="dexterity-types"
        )
        context = context.publishTraverse(self.request, name)

        factory = data.get("factory", None)
        if not factory:
            raise BadRequest("Missing parameter: 'factory'")

        if factory == "fieldset":
            res = add_fieldset(context, self.request, data)
        else:
            res = add_field(context, self.request, data)

        self.request.response.setStatus(201)
        return res



from AccessControl.SecurityManagement import getSecurityManager
from Acquisition import aq_base
from Acquisition.interfaces import IAcquirer
from base64 import b64decode
from email.utils import formatdate
from fnmatch import fnmatch
from plone.rest.interfaces import ICORSPolicy
from plone.restapi.exceptions import DeserializationError
from plone.restapi.interfaces import IDeserializeFromJson
from plone.restapi.services import Service
from plone.restapi.services.content.utils import add
from plone.restapi.services.content.utils import create
from plone.rfc822.interfaces import IPrimaryFieldInfo
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import base_hasattr
from Products.CMFPlone.utils import safe_hasattr
from uuid import uuid4
from zExceptions import Unauthorized
from zope.component import queryMultiAdapter
from zope.event import notify
from zope.interface import implementer
from zope.lifecycleevent import ObjectCreatedEvent
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces import NotFound

import json
import os
import time


TUS_OPTIONS_RESPONSE_HEADERS = {
    "Tus-Resumable": "1.0.0",
    "Tus-Version": "1.0.0",
    "Tus-Extension": "creation,expiration",
}


class UploadOptions(Service):
    """TUS upload endpoint for handling OPTIONS requests without CORS."""

    def reply(self):
        for name, value in TUS_OPTIONS_RESPONSE_HEADERS.items():
            self.request.response.setHeader(name, value)
        return super().reply()


class TUSBaseService(Service):
    def __call__(self):
        # We need to add additional TUS headers if this is a CORS preflight
        # request.
        policy = queryMultiAdapter((self.context, self.request), ICORSPolicy)
        if policy is not None:
            if self.request._rest_cors_preflight:
                policy.process_preflight_request()
                # Add TUS options headers in addition to CORS headers
                for name, value in TUS_OPTIONS_RESPONSE_HEADERS.items():
                    self.request.response.setHeader(name, value)
                return
            else:
                policy.process_simple_request()
        else:
            if self.request._rest_cors_preflight:
                return

        return self.render()

    def check_tus_version(self):
        version = self.request.getHeader("Tus-Resumable")
        if version != "1.0.0":
            return False
        return True

    def unsupported_version(self):
        self.request.response.setHeader("Tus-Version", "1.0.0")
        self.request.response.setStatus(412)
        return {
            "error": {"type": "Precondition Failed", "message": "Unsupported version"}
        }

    def error(self, type, message, status=400):
        """
        Set a status code (400 is the default error in the TUS
        reference server implementation) and return a plone.restapi
        conform error body.
        """
        self.request.response.setStatus(status)
        return {"error": {"type": type, "message": message}}


class UploadPost(TUSBaseService):
    """TUS upload endpoint for creating a new upload resource."""

    def reply(self):
        if not self.check_tus_version():
            return self.unsupported_version()

        length = self.request.getHeader("Upload-Length", "")
        try:
            length = int(length)
        except ValueError:
            return self.error("Bad Request", "Missing or invalid Upload-Length header")

        # Parse metadata
        metadata = {}
        for item in self.request.getHeader("Upload-Metadata", "").split(","):
            key_value = item.split()
            if len(key_value) == 2:
                key = key_value[0].lower()
                value = b64decode(key_value[1]).decode("utf-8")
                metadata[key] = value
        metadata["length"] = length
        if self.__name__.endswith("@tus-replace"):
            metadata["mode"] = "replace"
        else:
            metadata["mode"] = "create"

        tus_upload = TUSUpload(uuid4().hex, metadata=metadata)

        self.request.response.setStatus(201)
        self.request.response.setHeader(
            "Location",
            f"{self.context.absolute_url()}/@tus-upload/{tus_upload.uid}",
        )
        self.request.response.setHeader("Upload-Expires", tus_upload.expires())
        self.request.response.setHeader("Tus-Resumable", "1.0.0")
        return super().reply()


@implementer(IPublishTraverse)
class UploadFileBase(TUSBaseService):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.uid = None

    def publishTraverse(self, request, name):
        if self.uid is None:
            self.uid = name
        else:
            raise NotFound(self, name, request)
        return self

    def tus_upload(self):
        if self.uid is None:
            return

        tus_upload = TUSUpload(self.uid)
        length = tus_upload.length()
        if length == 0:
            return

        return tus_upload

    def check_add_modify_permission(self, mode):
        sm = getSecurityManager()
        if mode == "create":
            if not sm.checkPermission("Add portal content", self.context):
                raise Unauthorized
        else:
            if not sm.checkPermission("Modify portal content", self.context):
                raise Unauthorized


class UploadHead(UploadFileBase):
    """TUS upload endpoint for handling HEAD requests"""

    def reply(self):

        tus_upload = self.tus_upload()
        if tus_upload is None:
            return self.error("Not Found", "", 404)

        metadata = tus_upload.metadata()
        self.check_add_modify_permission(metadata.get("mode", "create"))

        if not self.check_tus_version():
            return self.unsupported_version()

        self.request.response.setHeader("Upload-Length", f"{tus_upload.length()}")
        self.request.response.setHeader("Upload-Offset", f"{tus_upload.offset()}")
        self.request.response.setHeader("Tus-Resumable", "1.0.0")
        self.request.response.setHeader("Cache-Control", "no-store")
        self.request.response.setStatus(200, lock=1)
        return super().reply()


@implementer(IPublishTraverse)
class UploadPatch(UploadFileBase):
    """TUS upload endpoint for handling PATCH requests"""

    def reply(self):

        tus_upload = self.tus_upload()
        if tus_upload is None:
            return self.error("Not Found", "", 404)

        metadata = tus_upload.metadata()
        self.check_add_modify_permission(metadata.get("mode", "create"))

        if not self.check_tus_version():
            return self.unsupported_version()

        content_type = self.request.getHeader("Content-Type")
        if content_type != "application/offset+octet-stream":
            return self.error("Bad Request", "Missing or invalid Content-Type header")

        offset = self.request.getHeader("Upload-Offset", "")
        try:
            offset = int(offset)
        except ValueError:
            return self.error("Bad Request", "Missing or invalid Upload-Offset header")

        request_body = self.request._file
        if hasattr(request_body, "raw"):  # Unwrap io.BufferedRandom
            request_body = request_body.raw
        tus_upload.write(request_body, offset)
        offset = tus_upload.offset()

        if tus_upload.finished:
            self.create_or_modify_content(tus_upload)
        else:
            self.request.response.setHeader("Upload-Expires", tus_upload.expires())

        self.request.response.setHeader("Tus-Resumable", "1.0.0")
        self.request.response.setHeader("Upload-Offset", f"{offset}")
        return self.reply_no_content()

    def create_or_modify_content(self, tus_upload):
        metadata = tus_upload.metadata()
        filename = metadata.get("filename", "")
        content_type = metadata.get("content-type", "application/octet-stream")
        mode = metadata.get("mode", "create")
        fieldname = metadata.get("fieldname")

        if mode == "create":
            type_ = metadata.get("@type")
            if type_ is None:
                ctr = getToolByName(self.context, "content_type_registry")
                type_ = ctr.findTypeName(filename.lower(), content_type, "") or "File"

            obj = create(self.context, type_)
        else:
            obj = self.context

        if not fieldname:
            info = IPrimaryFieldInfo(obj, None)
            if info is not None:
                fieldname = info.fieldname
            elif base_hasattr(obj, "getPrimaryField"):
                field = obj.getPrimaryField()
                fieldname = field.getName()

        if not fieldname:
            return self.error("Bad Request", "Fieldname required", 400)

        # Acquisition wrap temporarily for deserialization
        temporarily_wrapped = False
        if IAcquirer.providedBy(obj) and not safe_hasattr(obj, "aq_base"):
            obj = obj.__of__(self.context)
            temporarily_wrapped = True

        # Update field with file data
        deserializer = queryMultiAdapter((obj, self.request), IDeserializeFromJson)
        if deserializer is None:
            return self.error(
                "Not Implemented",
                f"Cannot deserialize type {obj.portal_type}",
                501,
            )
        try:
            deserializer(data={fieldname: tus_upload}, create=mode == "create")
        except DeserializationError as e:
            return self.error("Deserialization Error", str(e), 400)

        if temporarily_wrapped:
            obj = aq_base(obj)

        if mode == "create":
            if not getattr(deserializer, "notifies_create", False):
                notify(ObjectCreatedEvent(obj))
            obj = add(self.context, obj)

        tus_upload.close()
        tus_upload.cleanup()
        self.request.response.setHeader("Location", obj.absolute_url())


class TUSUpload:

    file_prefix = "tus_upload_"
    expiration_period = 60 * 60
    finished = False

    def __init__(self, uid, metadata=None):
        self.uid = uid

        self.tmp_dir = os.environ.get("TUS_TMP_FILE_DIR")
        if self.tmp_dir is None:
            client_home = os.environ.get("CLIENT_HOME")
            self.tmp_dir = os.path.join(client_home, "tus-uploads")
        if not os.path.isdir(self.tmp_dir):
            os.makedirs(self.tmp_dir)

        self.filepath = os.path.join(self.tmp_dir, self.file_prefix + self.uid)
        self.metadata_path = self.filepath + ".json"
        self._metadata = None

        if metadata is not None:
            self.initalize(metadata)

        self._file = None

    def initalize(self, metadata):
        """Initialize a new TUS upload by writing its metadata to disk."""
        self.cleanup_expired()
        with open(self.metadata_path, "w") as f:
            json.dump(metadata, f)

    def length(self):
        """Returns the total upload length."""
        metadata = self.metadata()
        if "length" in metadata:
            return metadata["length"]
        return 0

    def offset(self):
        """Returns the current offset."""
        if os.path.exists(self.filepath):
            return os.path.getsize(self.filepath)
        return 0

    def write(self, infile, offset=0):
        """Write to uploaded file at the given offset."""
        mode = "wb"
        if os.path.exists(self.filepath):
            mode = "ab+"
        with open(self.filepath, mode) as f:
            f.seek(offset)
            while True:
                chunk = infile.read(2 << 16)
                if not chunk:
                    offset = f.tell()
                    break
                f.write(chunk)
        length = self.length()
        if length and offset >= length:
            self.finished = True

    def open(self):
        """Open the uploaded file for reading and return it."""
        if self._file is None or self._file.closed:
            self._file = open(self.filepath, "rb")
        return self._file

    def close(self):
        """Close the uploaded file."""
        if self._file is not None and not self._file.closed:
            self._file.close()

    def metadata(self):
        """Returns the metadata of the current upload."""
        if self._metadata is None:
            if os.path.exists(self.metadata_path):
                with open(self.metadata_path, "rb") as f:
                    self._metadata = json.load(f)
        return self._metadata or {}

    def cleanup(self):
        """Remove temporary upload files."""
        if os.path.exists(self.filepath):
            os.remove(self.filepath)
        if os.path.exists(self.metadata_path):
            os.remove(self.metadata_path)

    def cleanup_expired(self):
        """Cleanup unfinished uploads that have expired."""
        for filename in os.listdir(self.tmp_dir):
            if fnmatch(filename, "tus_upload_*.json"):
                metadata_path = os.path.join(self.tmp_dir, filename)
                filepath = metadata_path[:-5]
                if os.path.exists(filepath):
                    mtime = os.stat(filepath).st_mtime
                else:
                    mtime = os.stat(metadata_path).st_mtime

                if (time.time() - mtime) > self.expiration_period:
                    os.remove(metadata_path)
                    if os.path.exists(filepath):
                        os.remove(filepath)

    def expires(self):
        """Returns the expiration time of the current upload."""
        if os.path.exists(self.filepath):
            expiration = os.stat(self.filepath).st_mtime + self.expiration_period
        else:
            expiration = time.time() + self.expiration_period
        return formatdate(expiration, False, True)



from plone.restapi.services import Service
from plone.app.multilingual.interfaces import ITranslationLocator


class TranslationLocator(Service):
    """Get translation locator placements information"""

    def reply(self):
        target_language = self.request.form["target_language"]

        locator = ITranslationLocator(self.context)
        parent = locator(target_language)

        return {"@id": parent.absolute_url()}



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from Products.Five.browser import BrowserView
from zope.component import provideAdapter
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserRequest

import json
import transaction
import unittest


class InternalServerErrorView(BrowserView):
    def __call__(self):  # pragma: no cover
        from urllib.error import HTTPError

        raise HTTPError(
            "http://nohost/plone/internal_server_error",
            500,
            "InternalServerError",
            {},
            None,
        )


class TestErrorHandling(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.portal.invokeFactory("Document", id="document1")
        self.document = self.portal.document1
        self.document_url = self.document.absolute_url()
        self.portal.invokeFactory("Folder", id="folder1")
        self.folder = self.portal.folder1
        self.folder_url = self.folder.absolute_url()
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    @unittest.skip("Not working since we moved to plone.rest")
    def test_404_not_found(self):
        response = self.api_session.get("non-existing-resource")
        self.assertEqual(response.status_code, 404)
        self.assertEqual(
            response.headers.get("Content-Type"),
            "application/json",
            "When sending a GET request with Accept: application/json "
            + "the server should respond with sending back application/json.",
        )
        self.assertTrue(json.loads(response.content))
        self.assertEqual("NotFound", response.json()["type"])

    @unittest.skip("Not working since we moved to plone.rest")
    def test_401_unauthorized(self):
        response = self.api_session.get(self.document_url)
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.headers.get("Content-Type"),
            "application/json",
            "When sending a GET request with Accept: application/json "
            + "the server should respond with sending back application/json.",
        )
        self.assertTrue(json.loads(response.content))
        self.assertEqual("Unauthorized", response.json()["type"])

    @unittest.skip("Not working since we moved to plone.rest")
    def test_500_internal_server_error(self):
        provideAdapter(
            InternalServerErrorView,
            adapts=(Interface, IBrowserRequest),
            provides=Interface,
            name="internal_server_error",
        )
        import transaction

        transaction.commit()

        response = self.api_session.get("internal_server_error")

        self.assertEqual(response.status_code, 500)
        self.assertEqual(
            response.headers.get("Content-Type"),
            "application/json",
            "When sending a GET request with Accept: application/json "
            + "the server should respond with sending back application/json.",
        )
        self.assertTrue(json.loads(response.content))
        self.assertEqual("HTTPError", response.json()["type"])



from plone.restapi.exceptions import DeserializationError
from plone.restapi.interfaces import IDeserializeFromJson
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services import Service
from plone.restapi.services.locking import is_locked
from zope.component import queryMultiAdapter


class ContentPatch(Service):
    """Updates an existing content object."""

    def reply(self):

        if is_locked(self.context, self.request):
            self.request.response.setStatus(403)
            return dict(error=dict(type="Forbidden", message="Resource is locked."))

        deserializer = queryMultiAdapter(
            (self.context, self.request), IDeserializeFromJson
        )
        if deserializer is None:
            self.request.response.setStatus(501)
            return dict(
                error=dict(
                    message="Cannot deserialize type {}".format(
                        self.context.portal_type
                    )
                )
            )

        try:
            deserializer()
        except DeserializationError as e:
            self.request.response.setStatus(400)
            return dict(error=dict(type="DeserializationError", message=str(e)))

        prefer = self.request.getHeader("Prefer")
        if prefer == "return=representation":
            self.request.response.setStatus(200)

            serializer = queryMultiAdapter(
                (self.context, self.request), ISerializeToJson
            )

            serialized_obj = serializer()
            return serialized_obj

        return self.reply_no_content()



from plone.restapi.setuphandlers import install_pas_plugin
from Products.CMFCore.utils import getToolByName


def install_pas_plugin_in_zope_root(setup_context):
    """Install PAS plugin in Zope root"""
    portal = getToolByName(setup_context, "portal_url").getPortalObject()
    install_pas_plugin(portal)
